from pygna.elaborators import TableElaboration as tE
import pygna.reading_class as rc
import logging
import sys
import pandas as pd

import pygna.output as output


class Converters:
    """
    This class is wrap static methods that can be used to convert the data from a format to another. Please refer to each class method for the specific function
    """

    def __init__(self):
        super(Converters, self).__init__()

    @classmethod
    def convert_e2s(cls, geneset: pd.DataFrame, tsv_data: pd.DataFrame, entrez_col: str = "NCBI Gene ID",
                    symbol_col: str = "Approved symbol") -> list:
        """
        Method to convert the entrez genes to symbols

        :param tsv_data: the dataframe to work on
        :param symbol_col: the column containing the symbols
        :param entrez_col: the column containing the entrez ID
        :param geneset: column containing the entrez to convert
        :return: list containing the string names

        Example
        _______
        >>> gmt_data = rc.ReadGmt(".gmt", True).get_data()
        >>> converted = []
        >>> for k, d in gmt_data.items():
        >>>     converted[k] = Converters.convert_e2s(d["genes"], tsv_data,entrez_col, symbol_col)

        """
        logging.info("Converting Entrez ID -> Symbols")
        unknown_counter = 0
        geneset_symbol = []
        for i in geneset:
            name = tsv_data[tsv_data[entrez_col] == int(i)][symbol_col].values.tolist()
            if len(name) > 0:
                geneset_symbol.append(str(name[0]))
            else:
                unknown_counter += 1
                geneset_symbol.append("<" + i + ">")
        if unknown_counter > 0:
            logging.warning("%d/%d terms that couldn't be mapped" % (unknown_counter, len(geneset)))

        return geneset_symbol

    @classmethod
    def convert_s2e(cls, geneset: pd.DataFrame, tsv_data: pd.DataFrame, entrez_col: str = "NCBI Gene ID",
                    symbol_col: str = "Approved symbol") -> list:
        """
        Method to convert the genes symbols to entrez id.

        :param tsv_data: the dataframe to work on
        :param symbol_col: the column containing the symbols
        :param entrez_col: the column containing the entrez ID
        :param geneset: column containing the strings to convert
        :return: list containing the entrez names

        Example
        _______
         >>> gmt_data = rc.ReadGmt(".gmt", True).get_data()
         >>> converted = []
         >>> for k, d in gmt_data.items():
         >>>    converted[k] = gmt_data[k]["genes"] = Converters.convert_s2e(d["genes"], tsv_data, entrez_col, symbol_col)
        """
        logging.info("Converting Symbols -> Entrez ID")
        geneset_entrez = []
        unknown_counter = 0
        for i in geneset:
            if type(i) != str:
                i = str(i)
            i = i.upper()
            name = tsv_data[tsv_data[symbol_col].str.upper() == i][entrez_col].values.tolist()
            if len(name) > 0:
                geneset_entrez.append(str(int(name[0])))
            else:
                unknown_counter += 1
                geneset_entrez.append("<" + i + ">")
        if unknown_counter > 0:
            logging.warning("%d/%d terms that couldn't be mapped" % (unknown_counter, len(geneset)))

        return geneset_entrez

    @staticmethod
    def _gmt_output(gmt_data: dict, gmt_output_file: str) -> None:
        """
        Save the dictionary on a GMT file

        :param gmt_data: the dictionary containing the data
        :param gmt_output_file: the file to save the data
        """
        output.print_GMT(gmt_data, gmt_output_file)


class CsvToCsvEnriched(Converters):
    """
    Class that is used to add a column with the entrezID or Symbols to a CSV file
    """

    def __init__(self, csv_file: pd.DataFrame, conversion: str, original_name_col: str, new_name_col: str, geneset: str,
                 entrez_col: str, symbol_col: str, converter_map_filename: str = "entrez_name.tsv",
                 output_file: str = None):
        """
        :param csv_file: dataframe with the data
        :param conversion: could be "e2s"-> Entrez2Symbols or "s2e" -> Symbol2Entrez
        :param original_name_col:  the column where to find the information to convert
        :param new_name_col: the name of the column that is going to contain the information
        :param geneset: the geneset to convert
        :param converter_map_filename:  the path to the .tsv used to convert the genes name
        :param output_file: [optional] the path of the output file
        :param entrez_col: the name of the entrez column
        :param symbol_col:  the name of the symbol column
        """
        super().__init__()
        logging.info("Adding values to the CSV file")
        self.filename = csv_file
        self.conversion = conversion
        self.original_name_col = original_name_col
        self.new_name_col = new_name_col
        self.geneset = geneset
        self.converter_map_file = converter_map_filename
        self.output = output_file
        self.entrez_col = entrez_col
        self.symbol_col = symbol_col

        self.map_table = rc.ReadTsv(self.converter_map_file, pd_table=True).get_data()
        self.file_data = rc.ReadCsv(self.filename).get_data()
        self._clean_table()

        if self.conversion == "e2s":
            self.file_data[self.new_name_col] = \
                super().convert_e2s(self.file_data[self.original_name_col].values.tolist(),
                                    self.map_table, self.entrez_col, self.symbol_col)
        elif self.conversion == "s2e":
            self.file_data[self.new_name_col] = \
                super().convert_s2e(self.file_data[self.original_name_col].values.tolist(),
                                    self.map_table, self.entrez_col, self.symbol_col)
        else:
            logging.error("Conversion type not understood")

        if self.output:
            self._csv_output()

    def get_data(self) -> pd.DataFrame:
        """
        Return the conversion result

        :return: dataframe with the e2s or s2e added as column
        """
        return self.file_data

    def _csv_output(self) -> None:
        """
        Print the output to a csv
        """
        output_file = self.output
        self.filename.to_csv(output_file, index=False)

    def _clean_table(self) -> None:
        """
        Method to make all upper and clean the table from null values
        """
        self.map_table = self.map_table.fillna("0")
        self.map_table["Approved symbol"] = self.map_table["Approved symbol"].str.upper()
        self.map_table["Synonyms"] = self.map_table["Synonyms"].str.upper()
        self.map_table["Previous symbols"] = self.map_table["Previous symbols"].str.upper()


class CsvToGmt(Converters):
    """
    This Class converts a csv file to a GMT allowing to filter the elements using the values of one of the columns.
    The user can specify the column used to retrieve the name of the objects and the filter condition. The output
    can be either a GMT with the names of the genes that pass the filter or a csv with the whole filtered table,
    otherwise both can be created.
    """

    def __init__(self, input_file: str, setname: str, filter_column: str, alternative: str, threshold: float,
                 output_gmt: str = None, output_csv: str = None, name_column: str = "Unnamed: 0",
                 descriptor: str = None):
        """
        :param input_file: str, the csv file
        :param setname: str, the name of the set
        :param output_gmt: str, output GMT name
        :param output_csv: str, output csv name
        :param name_column: str, column with the names
        :param filter_column: str, column with the values to be filtered
        :param alternative: str, alternative to use for the filterK with "less" the filter is applied <threshold; otherwise >= threshold
        :param threshold: float, threshold for the filter
        :param descriptor: str, descriptor for the GMT file
        """
        super().__init__()
        self.input_file = input_file
        self.setname = setname
        self.output_gmt = output_gmt
        self.output_csv = output_csv
        self.name_column = name_column
        self.filter_column = filter_column
        self.alternative = alternative
        self.threshold = threshold
        self.descriptor = descriptor

        self.table = rc.ReadCsv(self.input_file, column_to_fill=self.name_column).get_data()
        self.table = self._elaborate()

        if self.output_gmt:
            self._process_gmt()

        if self.output_csv:
            self._csv_output()

    def _elaborate(self) -> pd.DataFrame:
        """
        This method performs cleaning and filtering of the table

        :return: dataframe representing the cleaned and filter table
        """
        table = tE.clean_table(self.table, self.filter_column)
        table = tE.filter_table(table, filter_column=self.filter_column, alternative=self.alternative,
                                threshold=self.threshold)
        return table

    def _process_gmt(self) -> None:
        """
        This method parse the results and save them in a GMT file
        """
        if self.descriptor is None:
            self.descriptor = self.input_file.split("/")[-1]
        gmt_dict = {self.setname: {}}
        gmt_dict[self.setname]["descriptor"] = self.descriptor
        gmt_dict[self.setname]["genes"] = []
        geneset = self.table.loc[:, self.name_column].values.tolist()

        logging.info("geneset=" + str(geneset))
        gmt_dict[self.setname]["genes"] = geneset

        if self.output_gmt.endswith(".gmt"):
            super()._gmt_output(gmt_dict, self.output_gmt)
        else:
            logging.error("specify gmt output")

    def _csv_output(self) -> None:
        """
        This method save the dataframe in a CSV file
        """

        if self.output_csv.endswith(".csv"):
            self.table.to_csv(self.output_csv, sep=",", index=False)
        else:
            logging.error("specify csv output")


class GmtToGmtEnriched(Converters):
    """
    This Class is used to convert a GMT file, adding information about the Entrez ID or the symbol
    """

    def __init__(self, gmt_file: str, output_gmt_file: str, conversion: str, entrez_col: str, symbol_col: str,
                 converter_map_filename: str = "entrez_name.tsv"):
        """
        :param gmt_file: the input GMT file path
        :param output_gmt_file: the output GMT file path
        :param conversion: could be "e2s"-> Entrez2Symbols or "s2e" -> Symbol2Entrez
        :param entrez_col: the name of the entrez column
        :param symbol_col: the name of the symbol column
        :param converter_map_filename: the path to the .tsv used to convert the genes name
        """
        super().__init__()
        self.gmt_file = gmt_file
        self.output_gmt_file = output_gmt_file
        self.conversion = conversion
        self.entrez_col = entrez_col
        self.symbol_col = symbol_col
        self.converter_map_filename = converter_map_filename

        self.gmt_data = rc.ReadGmt(self.gmt_file, True).get_data()
        self.tsv_data = rc.ReadTsv(self.converter_map_filename, pd_table=True).get_data()

        if self.conversion == "e2s":
            for k, d in self.gmt_data.items():
                self.gmt_data[k]["genes"] = super().convert_e2s(d["genes"], self.tsv_data,
                                                                self.entrez_col, self.symbol_col)
        elif self.conversion == "s2e":
            for k, d in self.gmt_data.items():
                self.gmt_data[k]["genes"] = super().convert_s2e(d["genes"], self.tsv_data,
                                                                self.entrez_col, self.symbol_col)
        else:
            logging.error("Conversion type not understood")
        super()._gmt_output(self.gmt_data, self.output_gmt_file)


class GroupGmt(Converters):
    """
    This function generates a GMT file of multiple setnames. From the table file, it groups the names in the
    group_col (the column you want to use to group them) and prints the genes in the name_col. Set the descriptor
    according to your needs
    """

    def __init__(self, input_table: str, output_gmt: str, name_col: str = "Gene", group_col: str = "Cancer",
                 descriptor: str = "cancer_genes"):
        """
        :param input_table: the filename path (it must be in .csv, .tsv or .txt format)
        :param output_gmt: the output GMT file path
        :param name_col: the name of the column to write the genes
        :param group_col:the name of the column to group
        :param descriptor: the descriptor to use
        """
        super().__init__()
        self.input_table = input_table
        self.output_gmt = output_gmt
        self.name_col = name_col
        self.group_col = group_col
        self.descriptor = descriptor

        if self.input_table.endswith(".csv"):
            self.table = rc.ReadCsv(self.input_table, use_cols=[self.name_col, self.group_col]).get_data()
        elif self.input_table.endswith("tsv") or self.input_table.endswith("txt"):
            self.table = rc.ReadCsv(self.input_table, sep="\t", use_cols=[self.name_col, self.group_col]).get_data()
        else:
            sys.exit('Pass correct input (csv/tsv/txt)')
        self.gmt_data = self._elaborate()
        super()._gmt_output(self.gmt_data, self.output_gmt)

    def _elaborate(self) -> dict:
        """
        This method elaborates the table and returns a dictionary with the grouped columns

        :return: with the genes and the descriptor
        """
        diz = {}
        for g, group in self.table.groupby([self.group_col]):
            if len(group) < 10:
                print('warning: %s has less than 10 genes' % g)
            diz[g] = {}
            diz[g]['genes'] = group[self.name_col].astype(str).values.tolist()
            diz[g]['descriptor'] = self.descriptor
        return diz



import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import logging
from pygna import output
from pygna.utils import YamlConfig
import pandas as pd

import random
import string
import seaborn as sns

import pygna.output as output


class BlockModel(object):

    def __init__(self, block_model_matrix, n_nodes: int = 10, nodes_percentage: list = None):
        """
        This class implements a block model reading and elaboration methods

        :param block_model_matrix: the matrix to be used as block model
        :param n_nodes: the number of nodes
        :param nodes_percentage: the percentage of nodes to use for the calculations, passed through a list for example [0.5, 0.5]
        """
        self.n_nodes = n_nodes
        self.nodes = ["N" + str(i) for i in range(n_nodes)]
        self.n_clusters = block_model_matrix.shape[0]
        self.graph = nx.Graph()
        self.bm = block_model_matrix
        self.nodes_in_block = False
        self.nodes_percentage = nodes_percentage
        self.cluster_dict = {}

    def set_nodes(self, nodes_names: list) -> None:
        """
        Set the nodes name of the block model

        :param nodes_names: the names list

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> nodes = list("A", "B", "C")
        >>> bm.set_nodes(nodes)
        """
        self.nodes = nodes_names
        self.n_nodes = len(nodes_names)

    def set_bm(self, block_model_matrix: pd.DataFrame) -> None:
        """
        Change block model matrix used in the class

        :param block_model_matrix: the block model matrix

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> bmm = pd.DataFrame(mydata_matrix)
        >>> bm.set_bm(bmm)
        """

        if block_model_matrix.shape[0] == self.n_clusters:
            self.bm = block_model_matrix
        else:
            logging.error("the block model is supposed to have %d clusters" % (self.n_clusters))

    def set_nodes_in_block_percentage(self, nodes_percentage: list) -> None:
        """
        Pass the percentage of nodes in each block as a list, for example [0.5, 0.5]

        :param nodes_percentage: percentage of the nodes

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> bm.set_nodes_in_block_percentage([0.5, 0.5])
        """
        self.nodes_percentage = nodes_percentage

    def set_nodes_in_block(self, nodes_in_block: int) -> None:
        """
        Set the nodes number in the block model

        :param nodes_in_block: the number of nodes in the block list

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> bm.set_nodes_in_block(1000)
        """
        self.nodes_in_block = nodes_in_block

    def create_graph(self) -> None:
        """
        Create a graph from the parameters passed in the constructor of the class

        Example
        _______
        >>> bm = BlockModel(np.array(config["BlockModel"]["matrix"]), n_nodes=config["BlockModel"]["n_nodes"], nodes_percentage=config["BlockModel"]["nodes_percentage"])
        >>> bm.create_graph()
        """
        reject = True
        logging.info('Reject=' + str(reject))
        while reject:
            graph = generate_graph_from_sm(self.n_nodes, self.bm, self.nodes_in_block, self.nodes,
                                           self.nodes_percentage)
            LCC = max(nx.connected_components(graph), key=len)
            reject = (len(LCC) != self.n_nodes)
            logging.info('Reject=' + str(reject))
            logging.info('Nodes: %d, in LCC: %d' % (self.n_nodes, len(LCC)))

        self.graph = graph

    def plot_graph(self, output_folder: str) -> None:
        """
        Plot the block model graph

        :param output_folder: the folder where to save the result

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> bm.plot_graph("block_model_path.pdf")
        """
        plot_bm_graph(self.graph, self.bm, output_folder=output_folder)

    def write_network(self, output_file: str) -> None:
        """
        Save the network on a given file

        :param output_file: the output path where to save the results

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> bm.write_network("network.tsv")
        """
        self.network_file = output_file

        logging.info("Network written on %s" % (output_file))

        if output_file.endswith(".tsv"):
            nx.write_edgelist(self.graph, output_file, data=False, delimiter="\t")
        else:
            logging.error("output file format unknown")

    def write_cluster_genelist(self, output_file: str) -> None:
        """
        Save the gene list to a GMT file

        :param output_file: the output path where to save the results

        Example
        _______
        >>> p = 0.5
        >>> n_nodes = 1000
        >>> matrix = np.array([[1, 2], [3, 4]])
        >>> bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])
        >>> bm.write_cluster_genelist("genes.gmt")
        """
        self.genelist_file = output_file
        clusters = nx.get_node_attributes(self.graph, "cluster")

        for i in set(clusters.values()):
            c = "cluster_" + str(i)
            self.cluster_dict[c] = {}
            self.cluster_dict[c]["descriptor"] = "cluster"
            self.cluster_dict[c]["genes"] = [str(j) for j in clusters.keys() if clusters[j] == i]

        if output_file.endswith(".gmt"):
            output.print_GMT(self.cluster_dict, self.genelist_file)
        else:
            logging.error("output file format unknown")


def generate_graph_from_sm(n_nodes: int, block_model: pd.DataFrame, nodes_in_block: list = False,
                           node_names: list = None, nodes_percentage: list = None) -> nx.Graph:
    """
    This function creates a graph with n_nodes number of vertices and a matrix block_model that describes the intra e inter-block connectivity.
    The nodes_in_block is parameter, list, to control the number of nodes in each cluster

    :param n_nodes: the number of nodes in the block model
    :param block_model: the block model to elaborate
    :param nodes_in_block: the list of nodes in the block model
    :param node_names: the list of names in the block model
    :param nodes_percentage: the percentage of nodes to use for the calculations, passed through a list for example [0.5, 0.5]

    Example
    _______
    >>> bm = pd.DataFrame(mydata_matrix)
    >>> nodes = list("A","B","C")
    >>> graph = generate_graph_from_sm(n_nodes, bm, nodes_in_block, nodes, nodes_percentage)
    """

    if not node_names:
        node_names = range(n_nodes)

    edges = []
    G = nx.Graph()

    if nodes_percentage:
        cluster = np.random.choice(block_model.shape[0], size=n_nodes, p=nodes_percentage)
        np.random.shuffle(cluster)
    elif nodes_in_block:
        list_temp = [nodes_in_block[i] * [i] for i in range(len(nodes_in_block))]
        cluster = np.array([val for sublist in list_temp for val in sublist])
        np.random.shuffle(cluster)
    else:
        # cluster is an array of random numbers corresponding to the cluster of each node
        cluster = np.random.randint(block_model.shape[0], size=n_nodes)
    for i in range(n_nodes):
        G.add_node(node_names[i], cluster=cluster[i])

    for i in range(n_nodes):
        for j in range(i + 1, n_nodes):
            if np.random.rand() < block_model[cluster[i], cluster[j]]:
                edges.append((node_names[i], node_names[j]))

    G.add_edges_from(edges)
    return G


def plot_bm_graph(graph: nx.Graph, block_model: pd.DataFrame, output_folder: str = None) -> None:
    """
    Save the graph on a file

    :param graph: the graph with name of the nodes
    :param block_model: the block model
    :param output_folder: the folder where to save the file

    Example
    _______
    >>> bm = pd.DataFrame(mydata_matrix)
    >>> graph = nx.complete_graph(100)
    >>> plot_bm_graph(graph, bm, output_folder="./results/")
    """
    nodes = graph.nodes()
    colors = ['#b15928', '#1f78b4', '#6a3d9a', '#33a02c', '#ff7f00']
    cluster = nx.get_node_attributes(graph, 'cluster')
    labels = [colors[cluster[n]] for n in nodes]
    layout = nx.spring_layout(graph)

    plt.figure(figsize=(13.5, 5))
    plt.subplot(1, 3, 1)
    nx.draw(graph, nodelist=nodes, pos=layout, node_color='#636363', node_size=50, edge_color='#bdbdbd')
    plt.title("Observed network")

    plt.subplot(1, 3, 2)
    plt.imshow(block_model, cmap='OrRd', interpolation='nearest')
    plt.title("Stochastic block matrix")

    plt.subplot(1, 3, 3)
    legend = []
    for ix, c in enumerate(colors):
        legend.append(mpatches.Patch(color=c, label='C%d' % ix))

    nx.draw(graph, nodelist=nodes, pos=layout, node_color=labels, node_size=50, edge_color='#bdbdbd')
    plt.legend(handles=legend, ncol=len(colors), mode="expand", borderaxespad=0)
    plt.title("SB clustering")

    plt.savefig(output_folder + 'block_model.pdf', bbox_inches='tight')


def generate_sbm_network(input_file: "yaml configuration file") -> None:
    """
    This function generates a simulated network, using the block model matrix given as input and saves both the network and the cluster nodes.
    All parameters must be specified in a yaml file.
    This function allows to create network and geneset for any type of SBM
    """
    ym = YamlConfig()
    config = ym.load_config(input_file)
    print(config)

    bm = BlockModel(np.array(config["BlockModel"]["matrix"]), n_nodes=config["BlockModel"]["n_nodes"],
                    nodes_percentage=config["BlockModel"]["nodes_percentage"])
    outpath = config["Simulations"]["output_folder"]
    suffix = config["Simulations"]["suffix"]

    for i in range(config["Simulations"]["n_simulated"]):
        bm.create_graph()
        bm.write_network(outpath + suffix + "_s_" + str(i) + "_network.tsv")
        bm.write_cluster_genelist(outpath + suffix + "_s_" + str(i) + "_genes.gmt")
        # bm.plot_graph(outpath+suffix+"_s_"+str(i))


def generate_sbm2_network(output_folder: 'folder where the simulations are saved',
                          prefix: 'prefix for the simulations' = 'sbm',
                          n_nodes: 'nodes in the network' = 1000,
                          theta0: 'probability of connection in the cluster' = '0.9,0.7,0.5,0.2',
                          percentage: 'percentage of nodes in cluster 0, use ratio 0.1 = 10 percent' = '0.1',
                          density: 'multiplicative parameter used to define network density' = '0.06,0.1,0.2',
                          n_simulations: 'number of simulated networks for each configuration' = 3
                          ):
    """
    This function generates the simulated networks and genesets using the stochastic block model with 2 BLOCKS as described in the paper. The output names are going to be prefix_t_<theta0>_p_<percentage>_d_<density>_s_<n_simulation>_network.tsv or _genes.gmt
    One connected cluster while the rest of the network has the same probability of connection. SBM = d *[theta0, 1-theta0 1-theta0, 1-theta0]
    The simulator checks for connectedness of the generated network, if the generated net is not connected, a new simulation is generated.
    """

    teta_ii = [float(i) for i in theta0.replace(' ', '').split(',')]
    percentages = [float(i) for i in percentage.replace(' ', '').split(',')]
    density = [float(i) for i in density.replace(' ', '').split(',')]
    n_simulated = int(n_simulations)
    n_nodes = int(n_nodes)

    for p in percentages:
        for t in teta_ii:
            for d in density:
                matrix = np.array([[d * t, d * (1 - t)], [d * (1 - t), d * (1 - t)]])
                bm = BlockModel(matrix, n_nodes=n_nodes, nodes_percentage=[p, 1 - p])

                for i in range(n_simulated):
                    name = output_folder + prefix + "_t_" + str(t) + "_p_" + str(p) + "_d_" + str(d) + "_s_" + str(i)
                    bm.create_graph()
                    bm.write_network(name + "_network.tsv")
                    bm.write_cluster_genelist(name + "_genes.gmt")





def write_network(network, output_file):

    network_file= output_file
    logging.info("Network written on %s" %(output_file))
    if output_file.endswith(".tsv"):
        nx.write_edgelist(network, output_file, data=False, delimiter="\t")
    else:
        logging.error("output file format unknown")



def get_mix_genesets(gmt_diz,
                    tups = [('positive_0', 'positive_1'),
                            ('positive_2', 'positive_3'),
                            ('null_4', 'null_5'),
                            ('null_6', 'null_7')],
                    perc = [4,6,10,12,88,90,94,96]):

    diz = {}
    for t in tups:
        a = gmt_diz[t[0]]['genes']
        b = gmt_diz[t[1]]['genes']
        for p in perc:
            name = t[0]+'_'+str(int(p))+'_'+t[1]+'_'+str(int(100-p))
            aa = np.random.choice(a, int(len(a)/100*p), replace = False)
            bb = np.random.choice(b, int(len(a)/100*int(100-p)), replace = False)
            tot = []
            for i in aa:
                tot.append(i)
            for i in bb:
                tot.append(i)
            diz[name]=tot

    return(diz)


#########################################################################
####### COMMAND LINE FUNCTIONS ##########################################
#########################################################################

def generate_gna_sbm( output_tsv: 'output_network',
                        output_gmt: 'output geneset filename, this contains only the blocks',
                        output_gmt2: 'mixture output geneset filename, this contains the mixture blocks'=None,
                        N:'number of nodes in the network' = 1000,
                        block_size:'size of the first 8 blocks' = 50,
                        d:'baseline probability of connection, p0 in the paper' = 0.06,
                        fc_cis:'positive within-block scaling factor for the probability of connection, Mii = fc_cis * d (alpha parameter in the paper)' = 2.,
                        fc_trans:'positive between-block scaling factor for the probability of connection, (beta parameter in the paper)' = .5,
                        pi : 'percentage of block-i nodes for the genesets made of block-i and block-j. Use symmetrical values (5,95),use string comma separated' = '4,6,10,12,88,90,94,96',
                        descriptor='crosstalk_sbm',
                        sbm_matrix_figure: 'shows the blockmodel matrix' = None):

    """
    This function generates benchmark network and geneset to test
    the crosstalk between two blocks.

    This function generates 4 blocks with d*fold change probability
    and other 4 blocks with d probability.
    The crosstalk is set both between the the first 4 blocks and the others.

    Make sure that 8*cluster_size < N

    """

    clusters = 8
    lc = N - (block_size*clusters)
    if lc < 1:
        logging.error('nodes are less than cluster groups')

    d =float(d)
    sizes = clusters*[block_size]
    sizes.append(lc)
    print(sizes)

    probs = d*np.ones((9,9))
    #pp = np.tril(d/100*(1+np.random.randn(ncluster+1,ncluster+1)))

    A = fc_cis*d
    B = d + fc_trans*(d*(fc_cis-1))


    probs[0,1] = B
    probs[2,3] = B

    probs[1,0] = B
    probs[3,2] = B

    probs[4,5] = B
    probs[6,7] = B

    probs[5,4] = B
    probs[7,6] = B

    probs[0,0] = A
    probs[1,1] = A
    probs[2,2] = A
    probs[3,3] = A

    if type(sbm_matrix_figure)==str:
        f,ax = plt.subplots(1)
        sns.heatmap(probs, ax = ax, cmap = 'YlOrRd', annot=True)
        f.savefig(sbm_matrix_figure)

    ncycle = 0
    k = 0
    while (k<N):
        g = nx.stochastic_block_model(sizes, probs)
        g = max(nx.connected_component_subgraphs(g), key=len)
        k = len(g)
        ncycle +=1
        if ncycle > 20:
            logging.error('density is too low')

    H = nx.relabel_nodes(g, lambda x:'n'+str(x))

    gmt_diz = {}
    nodes = list(H.nodes)
    for p,l in enumerate(H.graph['partition'][:-1]):
        if p<4:
            name = 'positive_'+str(p)
        else:
            name = 'null_'+str(p)

        ll = [nodes[i] for i in l]

        gmt_diz[name]={}
        gmt_diz[name]['genes']=ll
        gmt_diz[name]['descriptor']=descriptor



    if type(output_gmt2)==str:
        perc = [float(i) for i in pi.split(',')]
        logging.info('Generating mixes with perc = %s')
        gmt_diz2={}
        mix_dix = get_mix_genesets(gmt_diz, perc = perc)
        for name,i in mix_dix.items():
            gmt_diz2[name]={}
            gmt_diz2[name]['genes']=i
            gmt_diz2[name]['descriptor']=descriptor
        output.print_GMT(gmt_diz2, output_gmt2)

    write_network(H, output_tsv)
    output.print_GMT(gmt_diz, output_gmt)
    print('Generated'+output_tsv)



def generate_gnt_sbm( output_tsv: 'output network filename',
                        output_gmt: 'output geneset filename, this contains only the blocks',
                        N:'number of nodes in the network' = 1000,
                        block_size: 'size of the first 6 blocks'= 50,
                        d: 'baseline probability of connection, p0 in the paper' = 0.06,
                        fold_change:'positive within-block scaling factor for the probability of connection, Mii = fold_change * d (alpha parameter in the paper)' = 2.,
                        descriptor:'descriptor for the gmt file'='mixed_sbm'):

    """
    This function generates 3 blocks with d*fold_change probability
    and other 3 blocks with d probability.
    Make sure that 6*cluster_size < N
    """


    lc = N - (block_size*6)
    if lc < 1:
        logging.error('nodes are less than cluster groups')

    d =float(d)
    sizes = 6*[block_size]
    sizes.append(lc)
    print(sizes)

    probs = d*np.ones((7,7))
    #pp = np.tril(d/100*(1+np.random.randn(ncluster+1,ncluster+1)))
    probs[0,0]=fold_change*d
    probs[1,1]=fold_change*d
    probs[2,2]=fold_change*d

    ncycle = 0
    k = 0
    while (k<N):
        g = nx.stochastic_block_model(sizes, probs)
        g = max(nx.connected_component_subgraphs(g), key=len)
        k = len(g)
        ncycle +=1
        if ncycle > 20:
            logging.error('density is too low')

    H = nx.relabel_nodes(g, lambda x:'n'+str(x))

    gmt_diz = {}
    nodes = list(H.nodes)
    for p,l in enumerate(H.graph['partition'][:-1]):
        if p<3:
            name = 'positive_'+str(p)
        else:
            name = 'null_'+str(p)

        ll = [nodes[i] for i in l]

        gmt_diz[name]={}
        gmt_diz[name]['genes']=ll
        gmt_diz[name]['descriptor']=descriptor

    write_network(H, output_tsv)
    output.print_GMT(gmt_diz, output_gmt)




import argh
import logging
import networkx as nx
import pygna.reading_class as rc
import pygna.output as out
import pygna.statistical_test as st
import pygna.painter as paint
import pygna.diagnostic as diagnostic
import pygna.command as cmd
import numpy as np


def average_closeness_centrality(graph: nx.Graph, geneset: set, diz: dict, observed_flag: bool = False) -> float:
    """
    This function calculates the average closeness centrality of a geneset. For
    a single node, the closeness centrality is defined as the inverse of the
    shortest path distance of the node from all the other nodes.

    Given a network with N nodes and a distance shortest path function between
    two nodes d(u,v) closeness centrality (u)= (N -1) / sum (v != u) d(u,v)

    :param graph: The network to analyse
    :param geneset: the geneset to analyse
    :param diz: The dictionary contains the nodes and a vector with the sum of shortest path for the whole network.
    """

    graph_centrality = []

    ids = [diz["nodes"].index(n) for n in geneset]
    graph_centrality = [(len(diz["nodes"]) - 1) / diz['vector'][idx] for idx in ids]

    return np.mean(graph_centrality)



def test_topology_centrality(
    network_file: "network file",
    geneset_file: "GMT geneset file",
    distance_matrix_filename: "The matrix with the SP for each node",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    cores: "Number of cores for the multiprocessing" = 1,
    in_memory: 'load hdf5 data onto memory' = False,):

    """
    This function calculates the average closeness centrality of a geneset.
    For a single node, the closeness centrality is defined as the inverse
    of the shortest path distance of the node from all the other nodes.

    """

    logging.info("Evaluating the test topology total degree, please wait")
    network = rc.ReadTsv(network_file).get_network()
    network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))
    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)
    setnames = [key for key in geneset.keys()]


    diz = {"nodes": cmd.read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[0],
           "matrix": cmd.read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[1]}
    diz["matrix"] = diz["matrix"] + np.transpose(diz["matrix"])

    np.fill_diagonal(diz["matrix"], float(0))

    diz['vector'] = np.sum(diz["matrix"],axis = 0)

    # Generate output
    output1 = out.Output(network_file, output_table, "topology_centrality", geneset_file, setnames)
    logging.info("Results file = " + output1.output_table_results)
    # Create table
    output1.create_st_table_empirical()
    st_test = st.StatisticalTest(average_closeness_centrality, network, diz)

    for setname, item in geneset.items():
        # Geneset smaller than size cut are not taken into consideration
        if len(item) > size_cut:
            item = set(item)
            observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item,
                                                                                     max_iter=number_of_permutations,
                                                                                     alternative="greater",
                                                                                     cores=cores)
            logging.info("Setname:" + setname)
            if n_mapped < size_cut:
                logging.info("%s removed from results since nodes mapped are < %d" % (setname, size_cut))
            else:
                logging.info("Observed: %g p-value: %g" % (observed, pvalue))
                output1.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed,
                                                  pvalue, np.mean(null_d), np.var(null_d))

    output1.close_temporary_table()


    logging.info("Test topology CENTRALITY completed")


def main():
    """
    argh dispatch
    """
    argh.dispatch_commands([test_topology_centrality])


if __name__ == "__main__":
    """
    MAIN
    """
    main()



import logging
import networkx as nx
import numpy as np
import scipy
import sys
import matplotlib.pyplot as plt
import pygna.reading_class as rc
import pygna.diagnostic as diag
import multiprocessing
import time


class KSTest:
    def __init__(self, distribution, network, diz={}):
        self.__distribution = distribution

        self.__network = network
        self.__diz = diz

        # print(type(self.__network))
        if (type(self.__network) is nx.Graph) or (type(self.__network) is nx.DiGraph):
            self.__universe = set(self.__network.nodes())
        elif type(self.__network) is dict:
            self.__universe = set(self.__network.keys())
            # print (self.__network [list(self.__network.keys())[0]])
        else:
            logging.error("Unknown network type: %s" % type(self.__network))
            sys.exit(-1)

    def apply_test(self, geneset):
        # mapping geneset
        mapped_geneset = sorted(list(set(geneset).intersection(self.__universe)))
        others = list(self.__universe.difference(set(mapped_geneset)))
        if mapped_geneset == 0:
            return 0, 0
        else:
            logging.info(
                "Mapped %d genes out of %d." % (len(mapped_geneset), len(geneset))
            )
            observed = self.__distribution(
                self.__network, mapped_geneset, self.__diz, observed_flag=True
            )
            baseline = self.__distribution(
                self.__network, others, self.__diz, observed_flag=True
            )
            logging.info("Observed Distribution evaluated")
            # iterations, the null distribution is a vector where ones represent rejected hypothesis
            stats, pvalue = scipy.stats.ks_2samp(observed, baseline, alternative='greater')
            # computing empirical pvalue

            return stats, pvalue, len(mapped_geneset), len(geneset)


###############################################################################
###  TEST STATISTICS FOR SINGLE GENESET  ######################################
###############################################################################


def degree_distribution(network, geneset, diz={}, observed_flag=False):
    """ degree distribution of the geneset """
    degree = nx.degree(network)
    distribution = np.array([degree[g] for g in geneset])
    return distribution




################################################################################
######### Degree Distribution test #############################################
################################################################################

# FIXME: problem with scipy version 1.2.1 , does not have alternative for test
def test_degree_distribution(
    network_file: "network file",
    geneset_file: "GMT geneset file",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    results_figure: "barplot of results, use pdf or png extension" = None,
    ):
    """
        Performs degree distribution test.
        Kolmogorov-Smirnov statistic on 2 samples.
        H0 is that the geneset is drawn from the same distribution of all the other nodes.
        H0 rejected if statistic is greater.
    """
    network = rc.ReadTsv(network_file)

    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)

    setnames = [key for key in geneset.keys()]

    output1 = out.Output(
        network_file, output_table, "test_degree", geneset_file, setnames
    )
    logging.info("Results file = " + output1.output_table_results)
    output1.create_st_table_empirical()

    st_test = KS.KSTest(KS.degree_distribution, network)

    for setname, item in geneset.items():

        item = set(item)
        if len(item) > size_cut:
            observed, pvalue, n_mapped, n_geneset = st_test.apply_test(item)
            logging.info("Setname:" + setname)
            if n_mapped < size_cut:
                logging.info(
                    "%s remove from results since nodes mapped are < %d"
                    % (setname, size_cut)
                )
            else:
                logging.info("Observed: %g p-value: %g" % (observed, pvalue))

                output1.update_st_table_empirical(
                    setname, n_mapped, n_geneset, 1, observed, pvalue, 0, 0
                )

    output1.close_temporary_table()
    if results_figure:
        paint.paint_datasets_stats(output1.output_table_results, results_figure, alternative='greater')



#TODO: painter class to be removed
class Painter:
    def __init__(self, network, output_folder, name, diz={}):

        self.network = network
        self.output_folder = output_folder
        self.name = name
        self.diz = diz

        if (type(self.network) is nx.Graph) or (type(self.network) is nx.DiGraph):
            self.universe = set(self.network.nodes())
        elif type(self.network) is dict:
            self.universe = set(self.network.keys())
        else:
            logging.error("Unknown network type: %s" % type(self.network))
            sys.exit(-1)


class Painter_RW(Painter):
    def __init__(self, network, output_folder, name, RW_dict, geneset, diz={}):
        Painter.__init__(self, network, output_folder, name, diz={})
        self.RW_nodes = list(RW_dict["nodes"])
        self.geneset = list(geneset)
        self.geneset_index = [list(RW_dict["nodes"]).index(i) for i in geneset]
        self.RW_matrix = RW_dict["matrix"][self.geneset_index, :][:, self.geneset_index]

    def plot_matrix(self, show_labels=False):

        logging.info("Plotting figure as " + str(self.output_folder + self.name))

        fig, axes = plt.subplots(1, 1)
        fig.subplots_adjust(left=0.2, right=0.99, bottom=0.2, top=0.99)
        axes.set_title("Diffusion matrix of geneset's nodes")
        mask = np.eye(self.RW_matrix.shape[0])

        logging.info(
            str(np.array(self.geneset)[:, np.newaxis].shape) + str(self.RW_matrix.shape)
        )

        if show_labels:
            tab_conversion = pd.read_table(
                "/home/viola/Desktop/geneset-network-analysis/primary_data/entrez_name.tsv",
                sep="\t",
            )
            labels = []
            for i in self.geneset:
                name = tab_conversion[tab_conversion["EntrezID"] == int(i)][
                    "Symbol"
                ].values.tolist()
                if len(name) > 0:
                    labels.append(str(name[0]))
                else:
                    labels.append(i)

            g2 = sns.heatmap(
                self.RW_matrix - mask,
                cmap="OrRd",
                square=True,
                ax=axes,
                xticklabels=labels,
                yticklabels=labels,
                cbar=True,
                vmin=0,
                vmax=np.max(self.RW_matrix - mask),
            )
            g2.set_yticklabels(g2.get_yticklabels(), rotation=0, fontsize=6)
            g2.set_xticklabels(g2.get_xticklabels(), rotation=90, fontsize=6)
        else:
            g2 = sns.heatmap(self.RW_matrix, cmap="OrRd", mask=mask, ax=axes, cbar=True)

        fig.savefig(
            self.output_folder + "RW_matrix_" + self.name + ".pdf", format="pdf"
        )

    def restrict_network(self, geneset):

        self.__network = nx.DiGraph(
            self.__network.subgraph(max(nx.connected_components(self.__network), key=len))
        )

    def add_annotation_nodes_in_geneset(self, geneset, annotation="in_subset"):

        logging.info("adding annotation to nodes in geneset")
        logging.info("%d genes in geneset,%d mapped in network ",
            (len(geneset), len(self.__universe.intersection(set(geneset)))),
        )
        geneset = self.__universe.intersection(set(geneset))
        dict_nodes = {}
        for n in self.__network.nodes():
            if n in geneset[setname]:
                dict_nodes[n] = True
            else:
                dict_nodes[n] = False
        nx.set_node_attributes(self.__network, annotation, dict_nodes)

    def add_weights_edges(self, dict_edges):
        pass

    def draw_graphml(self):

        """
        Draws a .graphml file from the network
        """
        # mapping geneset
        logging.info("Drawing graph for " + str(name))
        nx.write_graphml(
            self.__network, self.__output_folder + self.__name + ".graphml"
        )



import logging
import networkx as nx
import numpy as np
import sys
import multiprocessing
import pygna.statistical_test as st
from pygna.utils import get_sampling_p, get_node_bins_map

class StatisticalComparison:
    """
    This class implements the statistical analysis comparison between two genesets.
    Please refer to the single method documentation for the returning values
    """

    def __init__(self, comparison_statistic, network: nx.Graph, n_proc: int = 1, diz: dict = {}, degree_bins = 1):
        """
        :param comparison_statistic: the method do be applied to perform the statistical analysis
        :param network: the network to be used during the analysis
        :param n_proc: the number of CPU used to perform the elaboration
        :param diz: the dictionary used for the genesets
        """
        self.__comparison_statistic = comparison_statistic
        self.__network = network
        self.__diz = diz
        self.__n_proc = n_proc
        self.degree_bins = degree_bins
        self.node_2_bin_map = None
        self.bins = None

        if (type(self.__network) is nx.Graph) or (type(self.__network) is nx.DiGraph):
            self.__universe = list(set(self.__network.nodes()))
        elif type(self.__network) is dict:
            self.__universe = list(set(self.__network.keys()))

        else:
            logging.error("Unknown network type: %s" % type(self.__network))
            sys.exit(-1)

        # If number of bins is specified, the mapping is generated
        if (degree_bins>1):

            if type(self.__network) is dict:
                logging.error('Cannot do degree corrected sampling without the whole network')

            degree= list(self.__network.degree(nbunch = self.__universe))
            [nn,dd] = list(zip(*degree))

            # We make sure that later the universe corresponds to the mapping
            self.__universe = list(nn)
            histogram, bins_edge = np.histogram(list(dd), bins = degree_bins, density = False)
            self.bins = {int(i):{'nk':histogram[i], 'range':(bins_edge[i],bins_edge[i+1])} for i in range(len(histogram))}
            self.bins[len(self.bins.keys())-1]['range'] = (self.bins[len(self.bins.keys())-1]['range'][0], self.bins[len(self.bins.keys())-1]['range'][1]+1)
            self.node_2_bin_map = get_node_bins_map(degree, self.bins)

    def comparison_empirical_pvalue(self, genesetA: set, genesetB: set, alternative: str = "less", max_iter: int = 100,
                                    keep: bool = False) -> [int, float, float, int, int]:
        
        """
        Calculate the empirical value between two genesets

        :param genesetA: the first geneset to compare
        :param genesetB: the second geneset to compare
        :param alternative: the pvalue selection of the observed genes
        :param max_iter: the maximum number of iterations
        :param keep: if the geneset B should not be kept
        :return observed, pvalue, null_distribution, len(mapped_genesetA), len(mapped_genesetB): the list with the data calculated
        """

        # mapping genesets
        mapped_genesetA = sorted(list(set(genesetA) & set(self.__universe)))
        mapped_genesetB = sorted(list(set(genesetB) & set(self.__universe)))
        logging.info("Mapped %d genes out of %d from genesetA" % (len(mapped_genesetA), len(genesetA)))
        logging.info("Mapped %d genes out of %d from genesetB" % (len(mapped_genesetB), len(genesetB)))
        if (len(mapped_genesetA) == 0) or (len(mapped_genesetB) == 0):
            return 0, 1, np.array([0]), 0, 0
        else:

            sampling_p_a = None
            sampling_p_b = None
            if self.degree_bins > 1:
                # if more than one bin is specified, the sampling probability is generated
                sampling_p_a = get_sampling_p(mapped_genesetA, self.__network, self.bins, self.node_2_bin_map)
                if (~keep):
                    sampling_p_b = get_sampling_p(mapped_genesetB, self.__network, self.bins, self.node_2_bin_map)

            # Observed statistical values
            observed = self.__comparison_statistic(self.__network, mapped_genesetA, mapped_genesetB, self.__diz)

            # iterations
            null_distribution = StatisticalComparison.get_comparison_null_distribution_mp(self, mapped_genesetA,
                                                                                          mapped_genesetB, max_iter, keep,
                                                                                          sampling_p_a = sampling_p_a,
                                                                                          sampling_p_b = sampling_p_b,
                                                                                          )
            pvalue = 1
            if alternative == "greater":
                pvalue = (np.sum(null_distribution >= observed) + 1) / (float(len(null_distribution) + 1))
            else:
                pvalue = (np.sum(null_distribution <= observed) + 1) / (float(len(null_distribution) + 1))

            return observed, pvalue, null_distribution, len(mapped_genesetA), len(mapped_genesetB)

    def get_comparison_null_distribution_mp(self, genesetA: list, genesetB: list, max_iter: int = 100,
                                            keep: bool = False, 
                                            sampling_p_a = None, 
                                            sampling_p_b = None) -> np.ndarray:
        """
        Calculate the null distribution between two genesets with multiple CPUs

        :param genesetA: the first geneset to compare
        :param genesetB: the second geneset to compare
        :param max_iter: maximum number of iteration to perform
        :param keep: if the geneset B should not be kept
        :param sampling_p_a: random sampling probability for geneset a
        :param sampling_p_b: random sampling probability for geneset b
        :return: the array with null distribution
        """

        n_trial = int(max_iter / self.__n_proc)
        logging.info("n_proc = %d, each computing %d permutations " % (int(self.__n_proc), n_trial))

        # Don't set up the multicore architecture if only one core is given
        if self.__n_proc == 1:
            null_distribution = StatisticalComparison.get_comparison_null_distribution(self, genesetA, genesetB,
                                                                                       max_iter, keep,
                                                                                       sampling_p_a = sampling_p_a, 
                                                                                       sampling_p_b = sampling_p_b)
        else:
            p = multiprocessing.Pool(self.__n_proc)
            results = [p.apply_async(StatisticalComparison.get_comparison_null_distribution,
                                     args=(self, genesetA, genesetB, n_trial, keep, sampling_p_a, sampling_p_b)) for w in
                       list(range(1, self.__n_proc + 1))]

            null_distribution = np.array([])
            for r in results:
                null_distribution = np.hstack((null_distribution, np.array(r.get())))
            p.close()

        return np.asarray(null_distribution)

    def get_comparison_null_distribution(self, genesetA: list, genesetB: list, n_samples: int, keep: bool, 
                                            sampling_p_a = None, 
                                            sampling_p_b = None) -> list:
        """
        Calculate the null distribution between two genesets with single CPU

        :param genesetA: the first geneset to compare
        :param genesetB: the second geneset to compare
        :param n_samples: the number of samples to be taken
        :param keep: if the geneset B should not be kept
        :return: the random distribution calculated
        """
        np.random.seed()
        random_dist = []

        # association statistic, B kept, A bootstrapped
        if keep:
            for i in range(n_samples):
                random_sample_A = np.random.choice(self.__universe, len(genesetA), replace=False, p = sampling_p_a)
                random_dist.append(self.__comparison_statistic(self.__network, set(random_sample_A), set(genesetB),
                                                               self.__diz))
        # association statistic, B kept, A bootstrapped
        else:
            for i in range(n_samples):
                random_sample_A = np.random.choice(self.__universe, len(genesetA), replace=False, p = sampling_p_a)
                random_sample_B = np.random.choice(self.__universe, len(genesetB), replace=False, p = sampling_p_b)
                random_dist.append(self.__comparison_statistic(self.__network, set(random_sample_A),
                                                               set(random_sample_B), self.__diz))
        return random_dist


###############################################################################
###  TEST STATISTICS FOR COMPARISONS  #########################################
###############################################################################

def comparison_shortest_path(network: nx.Graph, genesetA: set, genesetB: set, diz: dict) -> float:
    """
    Evaluate the shortest path between two genesets

    :param network: the graph representing the network
    :param genesetA: the first geneset list
    :param genesetB: the second geneset list
    :param diz: the dictionary containing the nodes name and index
    """
    n = np.array([diz["nodes"].index(i) for i in genesetA])
    m = np.array([diz["nodes"].index(i) for i in genesetB])
    if len(n) == 0 or len(m) == 0:
        logging.info("Geneset length is equal to 0")
        sys.exit()
    cum_sum = calculate_sum(n, m, diz)
    cum_sum += calculate_sum(m, n, diz)
    d_AB = cum_sum / float(len(genesetA) + len(genesetB))
    d_A = st.geneset_localisation_statistic(network, genesetA, diz)
    d_B = st.geneset_localisation_statistic(network, genesetB, diz)
    return d_AB - (d_A + d_B) / 2


def calculate_sum(n: np.ndarray, m: np.ndarray, diz: dict) -> np.ndarray:
    """
    Evaluate the sum of the columns of two matrices

    :param n: the first column
    :param m: the second column
    :param diz: the dictionary containing the data
    """
    diz = diz["matrix"]
    sub_matrix = diz[n[:, None], m]
    sub_matrix = np.where(sub_matrix != np.inf, sub_matrix, 0)
    min_columns = np.amin(sub_matrix, axis=0)   
    sum_columns = np.sum(min_columns)
    return sum_columns


def comparison_random_walk(network: nx.Graph, genesetA: list, genesetB: list, diz: dict = {}) -> float:
    """
    Evaluate the random walk on two genesets

    :param network: the graph representing the network
    :param genesetA: the first geneset list
    :param genesetB: the second geneset list
    :param diz: the dictionary containing the nodes name and index
    """
    try:
        diz["matrix"]
    except KeyError:
        print("The dictionary doesnt have a matrix key")
        raise

    genesetA_index = [diz["nodes"].index(i) for i in genesetA]
    genesetB_index = [diz["nodes"].index(i) for i in genesetB]

    if len(genesetA_index) == 0 or len(genesetB_index) == 0:
        sys.exit()

    probAB = [diz["matrix"][i, j] for i in genesetA_index for j in genesetB_index]
    probBA = [diz["matrix"][i, j] for i in genesetB_index for j in genesetA_index]

    prob = np.sum(probAB) + np.sum(probBA)
    return prob



# geneset_network_analysis_tool : Python package

"""Top-level package for geneset_network_analysis_tool."""

__author__ = """Viola Fanfani & Giovanni Stracquadanio"""
__email__ = 'v.fanfani@sms.ed.ac.uk'
__version__ = '3.3.1'




import networkx as nx
import math
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import logging
from pygna import output


class DegreeModel(object):
    """
    This class is used to calculate a degree model
    """
    def __init__(self, network_prob: float = 0.5, vip_prob: float = 1, n_nodes: int = 10,
                 vip_percentage: float = 0.1):
        self.n_nodes = n_nodes
        self.graph = nx.Graph()
        self.network_prob = network_prob
        self.vip_prob = vip_prob
        self.n_vip = math.ceil(vip_percentage * self.n_nodes)
        self.nodes = ["N" + str(i) for i in range(n_nodes)]
        self.cluster_dict = {}

    def set_nodes(self, nodes_names: list) -> None:
        """
        Set the name of the nodes and calculate the lenght of the list

        :param nodes_names: the list with the nodes name

        Example
        _______
        >>> nodes = list("A", "B", "C")
        >>> dm = DegreeModel()
        >>> dm.set_nodes(nodes)
        """
        self.nodes = nodes_names
        self.n_nodes = len(nodes_names)

    def create_graph(self) -> None:
        """
        Create a graph from the nodes

        Example
        _______
        >>> dm = DegreeModel()
        >>> dm.create_graph()
        """
        reject = True
        logging.info("Reject=" + str(reject))
        while reject:
            graph = generate_graph_vip(self.n_nodes, self.n_vip, network_prob=self.network_prob, vip_prob=self.vip_prob,
                                       node_names=self.nodes)
            LCC = max(nx.connected_components(graph), key=len)
            reject = len(LCC) != self.n_nodes
            logging.info("Reject=" + str(reject))
            logging.info("Nodes: %d, in LCC: %d" % (self.n_nodes, len(LCC)))

        self.graph = graph


    def write_network(self, output_file: str) -> None:
        """
        Write on file the network as an edge list

        :param output_file: the file path where to save the network

        Example
        _______
        >>> dm = DegreeModel()
        >>> dm.write_network("myoutputfile.tsv")
        """
        self.network_file = output_file

        logging.info("Network written on %s" % output_file)

        if output_file.endswith(".tsv"):
            nx.write_edgelist(self.graph, output_file, data=False, delimiter="\t")
        else:
            logging.error("output file format unknown")

    def write_genelist(self, output_file: str) -> None:
        """
        Write the GMT gene list on a specific file

        :param output_file: the file path where to save the gene list

        Example
        _______
        >>> dm = DegreeModel()
        >>> dm.write_genelist("myoutput.gmt")
        """
        self.genelist_file = output_file

        clusters = nx.get_node_attributes(self.graph, "cluster")

        for i in set(clusters.values()):
            c = "cluster_" + str(i)
            self.cluster_dict[c] = {}
            self.cluster_dict[c]["descriptor"] = "cluster"
            self.cluster_dict[c]["genes"] = [
                str(j) for j in clusters.keys() if clusters[j] == i
            ]

        if output_file.endswith(".gmt"):
            output.print_GMT(self.cluster_dict, self.genelist_file)
        else:
            logging.error("output file format unknown")


def generate_graph_vip(n_nodes: int, n_vip: int, network_prob: float = 0.5, vip_prob: float = 1,
                       node_names: list = None) -> nx.Graph:
    """
    This function creates a graph with n_nodes number of vertices and a matrix block_model that describes the intra e inter-block connectivity.
    The nodes_in_block is parameter, list, to control the number of nodes in each cluster

    :param n_nodes: number of nodes in the network
    :param n_vip: number of VIP to create
    :param network_prob: probability of connection in the network
    :param vip_prob: probability of connection of the vip
    :param node_names: list of nodes for the network

    Example
    _______
    >>> n_nodes = 100
    >>> n_vip = 10
    >>> network_prob = 0.5
    >>> vip_prob = 1
    >>> nodes = ["A", "B", "C"]
    >>> graph = generate_graph_vip(n_nodes, n_vip, network_prob=network_prob, vip_prob=vip_prob, node_names=nodes)
    """

    if not node_names:
        node_names = range(n_nodes)

    edges = []
    G = nx.Graph()

    list_temp = [(n_nodes - n_vip) * [0]]
    list_temp.append(n_vip * [1])

    cluster = np.array([val for sublist in list_temp for val in sublist])
    np.random.shuffle(cluster)

    prob = [network_prob, vip_prob]
    p = [prob[i] for i in cluster]

    for i in range(n_nodes):
        G.add_node(node_names[i], cluster=cluster[i], prob=p[i])

    for i in range(n_nodes):
        for j in range(i + 1, n_nodes):
            if np.random.binomial(1, p[i]) == 1:
                edges.append((node_names[i], node_names[j]))

    G.add_edges_from(edges)
    return G


def plot_vip_graph(graph: nx.Graph, output_folder: str = None) -> None:
    """
    Plot the VIP graph in the specific folder

    :param graph: the graph to plot
    :param output_folder: the folder path where to save the file

    Example
    _______
    >>> g = nx.complete_graph(100)
    >>> plot_vip_graph(g, "./")
    """
    nodes = graph.nodes()
    colors = ["#b15928", "#1f78b4"]
    cluster = nx.get_node_attributes(graph, "cluster")
    labels = [colors[cluster[n]] for n in nodes]
    layout = nx.spring_layout(graph)

    plt.figure(figsize=(13.5, 5))
    plt.subplot(1, 3, 1)
    nx.draw(
        graph,
        nodelist=nodes,
        pos=layout,
        node_color="#636363",
        node_size=50,
        edge_color="#bdbdbd",
    )
    plt.title("Observed network")

    plt.subplot(1, 3, 2)
    plt.imshow(nx.adjacency_matrix(graph).toarray(), cmap="OrRd")
    plt.title("Adjacency Matrix")

    plt.subplot(1, 3, 3)
    legend = []
    for ix, c in enumerate(colors):
        legend.append(mpatches.Patch(color=c, label="C%d" % ix))

    nx.draw(
        graph,
        nodelist=nodes,
        pos=layout,
        node_color=labels,
        node_size=50,
        edge_color="#bdbdbd",
    )
    plt.legend(handles=legend, ncol=len(colors), mode="expand", borderaxespad=0)
    plt.title("VIP clustering")

    plt.savefig(output_folder + "VIP.pdf", bbox_inches="tight")


def plot_adjacency(graph: nx.Graph, output_folder: str, prefix: str) -> None:
    """
    Plot the adjacency matrix on file

    :param graph: the graph to plot
    :param output_folder: the folder where to save the file
    :param prefix: the prefix to give to the file

    Example
    _______
    >>> g = nx.complete_graph(100)
    >>> plot_adjacency(g, "./","adj_matrix_")
    """
    plt.figure(figsize=(13.5, 5))

    plt.subplot(1, 1, 1)
    plt.imshow(nx.adjacency_matrix(graph).toarray(), cmap="OrRd")
    plt.title("Adjacency Matrix")

    plt.savefig(output_folder + prefix + "VIP.png")


def generate_hdn_network(output_folder: 'the output folder path',
                        prefix:'the prefix of the file to be saved',
                        n_nodes: 'the number of nodes in the network' = 1000,
                        network_prob: 'probability of connection in the network' = 0.005,
                         hdn_probability: 'probability of connection of the HDNs'= 0.3,
                         hdn_percentage: 'percentage of HDNs' = 0.05,
                         number_of_simulations: int = 5):

    """
    This function generates a simulated network using the VIP model

    """

    dm = DegreeModel(
        network_prob=network_prob,
        vip_prob=hdn_probability,
        n_nodes=n_nodes,
        vip_percentage=hdn_percentage,
    )

    for i in range(number_of_simulations):
        dm.create_graph()
        dm.write_network(output_folder + prefix + "_s_" + str(i) + "_network.tsv")
        dm.write_genelist(output_folder + prefix + "_s_" + str(i) + "_genes.gmt")
        plot_adjacency(dm.graph, output_folder, prefix=prefix + "_s_" + str(i))




def hdn_add_branching(input_geneset_file:'input geneset containing the sets to be merged',
                    network_file: 'network filename',
                    hdn_set:'setname of the HDNs in the geneset'='cluster_1',
                    general_set:'other setname in the geneset'='cluster_0',
                    number_of_hdns:'number of seed HDNs to start adding nodes'=3,
                    number_of_reps: 'number of new genesets created for each condition'=3,
                    output_geneset_file:'if no output gmt filename is passed, the data is added to the input file'=None,
                    output_graph_file:'if graphml file is passed the network with labels is written' = None ):

    """
    Creates new genesets from the vip list, new genesets are created adding 1 step
    nodes to vips. The new genes are created as branches.
    """

    if output_geneset_file:
        logging.info('output written to: %s ' %output_geneset_file)
    else:
        output_geneset_file=input_geneset_file
        logging.info('output written on the same file as the input %s' %output_geneset_file)


    geneset = ps.GMTParser().read(input_geneset_file, read_descriptor=True)

    network = ps.__load_network(network_file)
    print(type(network))


    # If graphml output is set, we create a MultiGraph from the original network
    if output_graph_file:
        MG = nx.MultiDiGraph()
        MG.add_nodes_from(network)
        print(MG.nodes())
    try:
        vips=set(geneset[hdn_set]['genes'])
        general=set(geneset[general_set]['genes'])

    except:
        sys.exit('The vip setname could not be read')
    if output_graph_file:
        # add annotation
        vip_dict = { i:1 if i in vips else 0 for i in list(network.nodes()) }
        nx.set_node_attributes(MG, vip_dict, 'vip')

    # new_geneset={}
    # new_geneset[hdn_set]={}
    # new_geneset[hdn_set]['descriptor']='original clusters'
    # new_geneset[hdn_set]['genes']=list(vips)

    # new_geneset[general_set]={}
    # new_geneset[general_set]['descriptor']='original clusters'
    # new_geneset[general_set]['genes']=list(general)
    new_geneset=geneset

    # A geneset is created for each repetition and combination of
    # of first layer connections and length of the branch
    for rep in range(number_of_reps):
        for first_layer in [1]:
            for branch_length in [5, 10, 15]:

                new_name='branching_'+str(first_layer)+'_'+str(branch_length)+'_'+str(rep)
                new_set=[]

                for k in range(number_of_hdns):
                    seed=random.choice(list(vips))
                    new_set.append(seed)
                    for fl in range(first_layer):
                        # From the seed node, get a first connection node
                        seed_edges=[i[1] for i in network.edges(seed)]
                        new_node=random.choice(seed_edges)
                        while new_node in new_set:
                            new_node=random.choice(seed_edges)

                        new_set.append(new_node)
                        #if output_graph_file:
                        #    MG.add_edges_from([(seed,new_node,{'branch':new_name, 'number':0})])
                        #    nx.set_node_attributes(MG, {seed:'1'}, 'in_set')

                        for br in range(branch_length-1):
                            new_node_edges=[i[1] for i in network.edges(new_node)]
                            if len(set(new_node_edges).difference(set(new_set)))>0:
                                old_node=new_node
                                new_node=random.choice(list(set(new_node_edges).difference(set(new_set))))

                                new_set.append(new_node)


                                kkk=nx.shortest_path_length(network, source=old_node, target=new_node)
                                if kkk>1:
                                    print('WARNING path length: %d' %kkk)

                                if output_graph_file:
                                    MG.add_edges_from([(old_node,new_node,{'branch':new_name, 'number':br+1})])
                                    nx.set_node_attributes(MG, {old_node:'1'}, 'in_set')
                if output_graph_file:
                    nx.set_node_attributes(MG, {new_node:'1'}, 'in_set')
                new_geneset[new_name]={}
                new_geneset[new_name]['descriptor']='branching'
                new_geneset[new_name]['genes']=new_set

    all_nodes = set()
    for k,diz in new_geneset.items():
        if diz['descriptor']=='branching':

            all_nodes=all_nodes.union(set(diz['genes']) )
            for g in diz['genes']:
                all_nodes=all_nodes.union(set(network[g].keys()))
                ed=[(g,i) for i in network[g].keys()]
                if output_graph_file:
                    MG.add_edges_from(ed)


    universe=set(network.nodes())
    difference= universe.difference(all_nodes)

    out.print_GMT(new_geneset, output_geneset_file)

    if output_graph_file:
        MG.remove_nodes_from(difference)
        nx.write_graphml(MG, output_graph_file+ ".graphml")

def hdn_add_partial(input_geneset_file:'input geneset containing the sets to be merged',
                    hdn_set:'setname of the HDNs in the geneset'='cluster_1',
                    general_set:'other setname in the geneset'='cluster_0',
                    reps: 'number of new genesets made of part of vips' = 3,
                    percentage_partial_vips: 'percentage of HDNs in the new geneset' = '0.1,0.2,0.5',
                    output_geneset_file:'if no output gmt filename is passed, the data is added to the input file'=None):

    """
    Creates new genesets from the vip list, number of genesets and portion of
    genes can be specified by input.
    """

    if type(percentage_partial_vips)==str:
        percentage_partial_vips=percentage_partial_vips.replace('[','').replace(']','').replace(' ','')
        percentage_partial_vips=[float(i) for i in percentage_partial_vips.split(',')]

    if output_geneset_file:
        logging.info('output written to: %s ' %output_geneset_file)
    else:
        output_geneset_file=input_geneset_file
        logging.info('output written on the same file as the input %s' %output_geneset_file)

    geneset = ps.GMTParser().read(input_geneset_file, read_descriptor=True)

    try:
        vips=set(geneset[hdn_set]['genes'])
        general=set(geneset[general_set]['genes'])
    except:
        sys.exit('The vip setname could not be read')

    # Genesets made of part of the vips
    for perc in percentage_partial_vips:
        for rep in range(reps):
            new_name='Partial_vip_'+str(perc)+'_'+str(rep)
            new_set=np.random.choice(list(vips), int(perc*len(vips)))
            geneset[new_name]={}
            geneset[new_name]['descriptor']='partial vip list'
            geneset[new_name]['genes']=list(new_set)

    out.print_GMT(geneset, output_geneset_file)

def hdn_add_extended(input_geneset_file:'input geneset containing the sets to be merged',
                    hdn_set:'setname of the HDNs in the geneset'='cluster_1',
                    general_set:'other setname in the geneset'='cluster_0',
                    reps: 'number of new genesets made of part of HDNs' = 3,
                    percentage_extended_vips: 'percentage of HDNs in the new genesett' = '[0.2]',
                    ratio_others: 'ratio of genes to add to HDNs' = '[2,2.5,3,4]',
                    output_geneset_file:'if no output gmt filename is passed, the data is added to the input file'=None):

    """
    Creates new genesets from the vip list, number of genesets and portion of genes
    can be specified by input. The final new geneset is going to be formed by:
    percentage ev*HDN_total + ratio*percentage ev*vips total.

    """

    if type(percentage_extended_vips)==str:
        percentage_extended_vips=percentage_extended_vips.replace('[','').replace(']','').replace(' ','')
        percentage_extended_vips=[float(i) for i in percentage_extended_vips.split(',')]

    if type(ratio_others)==str:
        ratio_others=ratio_others.replace('[','').replace(']','').replace(' ','')
        ratio_others=[float(i) for i in ratio_others.split(',')]

    if output_geneset_file:
        logging.info('output written to: %s ' %output_geneset_file)
    else:
        output_geneset_file=input_geneset_file
        logging.info('output written on the same file as the input %s' %output_geneset_file)

    geneset = ps.GMTParser().read(input_geneset_file, read_descriptor=True)

    try:
        vips=set(geneset[hdn_set]['genes'])
        general=set(geneset[general_set]['genes'])
    except:
        sys.exit('The vip setname could not be read')

    # VIPs and other genes
    for rep in range(reps):
        for perc_ext in percentage_extended_vips:

            new_vips=np.random.choice(list(vips), int(perc_ext*len(vips)))
            for add in ratio_others:
                new_name='Extend_vip_'+str(perc_ext)+'_'+str(add)+'_'+str(rep)
                # Add a number of genes smaller than the total of generals
                new_genes_number=min(len(general), int(add*np.ceil(perc_ext*len(vips))) )
                new_genes=np.random.choice(list(general),new_genes_number )
                new_set=set(new_vips).union(set(new_genes))
                geneset[new_name]={}
                geneset[new_name]['descriptor']='extended vip and other genes list'
                geneset[new_name]['genes']=list(new_set)

    out.print_GMT(geneset, output_geneset_file)



import yaml
import pygna.converters as pc
import pygna.elaborators as pe


class YamlConfig:
    def __init__(self):
        pass

    def write_config(self, data, filename):
        with open(filename, "w") as outfile:
            yaml.dump(data, outfile, default_flow_style=True)

    def load_config(self, filename):
        with open(filename, "r") as stream:
            try:
                config = yaml.load(stream)
                return config

            except yaml.YAMLError as exc:
                return print(exc)


def convert_gmt(gmt_file: "gmt file to be converted",
                output_gmt_file: "output file",
                conversion: "e2s or s2e",
                converter_map_filename: "tsv table used to convert gene names",
                entrez_col: "name of the entrez column" = "NCBI Gene ID",
                symbol_col: "name of the symbol column" = "Approved symbol"):
    """
    This function is used to convert a GMT file, adding information about the Entrez ID or the symbol
    """
    pc.GmtToGmtEnriched(gmt_file, output_gmt_file, conversion, entrez_col, symbol_col, converter_map_filename)


def geneset_from_table(input_file: "input csv file",
                       setname: "name of the set",
                       output_gmt: "output gmt name" = None,
                       output_csv: "output csv name" = None,
                       name_column: "column with the names" = "Unnamed: 0",
                       filter_column: "column with the values to be filtered" = "padj",
                       alternative: "alternative to use for the filter, with less the filter is applied <threshold, "
                                    "otherwise >= threshold" = "less",
                       threshold: "threshold for the filter" = 0.01,
                       descriptor: "descriptor for the gmt file" = None):
    """
    This function converts a csv file to a GMT allowing to filter the elements using the values of one of the columns.
    The user can specify the column used to retrieve the name of the objects and the filter condition. The output
    can be either a GMT with the names of the genes that pass the filter or a csv with the whole filtered table,
    otherwise both can be created.
    """
    pc.CsvToGmt(input_file, setname, filter_column, alternative, threshold, output_gmt, output_csv, name_column,
                descriptor)


def filter_table(table: "input csv file",
                 filter_column: "column with the values to be filtered" = "padj",
                 alternative: "alternative to use for the filter, with less the filter is applied <threshold, "
                              "otherwise >= threshold" = "less",
                 threshold: "threshold for the filter" = 0.01):
    return pe.TableElaboration.filter_table(table, filter_column, alternative, threshold)


def generate_group_gmt(input_table: "table to get the geneset from",
                       output_gmt: "output gmt file",
                       name_col='Gene',
                       group_col='Cancer',
                       descriptor='cancer_genes'):
    """
    This function generates a GMT file of multiple setnames. From the table file, it groups the names in the
    group_col (the column you want to use to group them) and prints the genes in the name_col. Set the descriptor
    according to your needs
    """
    pc.GroupGmt(input_table, output_gmt, name_col, group_col, descriptor)


def convert_csv(csv_file: "csv file where to add a name column",
                conversion: "e2s or s2e",
                original_name_col: "column name to be converted",
                new_name_col: "name of the new column with the converted names",
                geneset: "the geneset to convert",
                converter_map_filename: "tsv table used to convert gene names" = "entrez_name.tsv",
                output_file: "if none, table is saved in the same input file" = None,
                entrez_col: "name of the entrez column" = "NCBI Gene ID",
                symbol_col: "name of the symbol column" = "Approved symbol"):
    """
    This function is used to add a column with the entrezID or Symbols to a CSV file
    """
    pc.CsvToCsvEnriched(csv_file, conversion, original_name_col, new_name_col, geneset, entrez_col, symbol_col,
                        converter_map_filename, output_file)


################################################################
########### Functions for corrected sampling ###################
################################################################

def get_bin(degree, map_bin, name=None):
    # For a node degree return the bin_name according to the ranges
    for r,i in map_bin.items():
        if (degree>=r[0]) & (degree<r[1]):
            return(i)

def get_node_bins_map(degree, bin_histogram):
    # Reshape the dictionary to have range:bin_number
    map_bin = {val['range']:i for i,val in bin_histogram.items()}
    # Get a map of the bin of each node node:bin_number
    node_2_bin = {d[0]:get_bin(d[1], map_bin) for d in degree}
    return node_2_bin

def get_sampling_p(mapped_geneset, network, bins, node_2_bin_map):
    # Reshape the dictionary to have range:bin_number
    map_bin = {val['range']:i for i,val in bins.items()}
    # Get the degree of the mapped geneset
    degree = network.degree(nbunch = mapped_geneset)
    # Get a map of the bin of each node node:bin_number
    node_2_bin = [get_bin(d[1], map_bin) for d in degree]
    # Count the mapped node for each bin
    counts = {k:node_2_bin.count(k) for k in bins.keys()}
    # Get the probability for each bin bin_number:probability
    # For each bin we have:
    #   - c = number of mapped nodes in the bin
    #   - nk = number of total nodes in the bin
    #   - m = total number of mapped nodes
    # p = c/nk/m, indeed c/m would be the probability of choosing the bin, 
    # c/nk/m becomes the probability of choosing one of the nodes in the bin
    bin_2_p_map = {(k):(counts[k]/val['nk']/len(mapped_geneset) if (val['nk']>0) else 0) for k,val in bins.items()}
    # Get the probability of sampling for each node
    p = [bin_2_p_map[int(k)] for n,k in node_2_bin_map.items()]
    return(p)


"""PyGNA commands main module
"""
import logging
import numpy as np
import pygna.output as out
import networkx as nx
import pygna.elaborators as pe
import pygna.reading_class as rc
import pygna.statistical_test as st
import pygna.statistical_comparison as sc
import pygna.diagnostic as diagnostic
import pygna.painter as paint
import pygna.utils as utils
import pygna.statistical_diffusion as sd
import scipy
import scipy.linalg.interpolative
import itertools
import tables
import mygene



def read_distance_matrix(distance_matrix_filename, in_memory=False):
    """
    Reads the large matrix. Uses a hdf5 file to work with it

    :param distance_matrix_filename: str, the file to read
    :param in_memory: bool, if the table must be kept in memory
    """
    nodes, data = rc.ReadDistanceMatrix(distance_matrix_filename, in_memory).get_data()
    return nodes, data


################################################################################
######### SUMMARY ##############################################################
################################################################################

def network_summary(network_file: "network file",
                    text_output: "output text file for the summary",
                    degree_figure_file: "pdf or png file for the degree distribution",
                    c_components_figure_file: "pdf or png file for the connected components distribution",
                    geneset_input_file: "geneset file" = None,
                    setname: "specify a single geneset" = None):
    '''This function saves the principal info of a graph.
    | - network properties \n
    - degree distribution\n
    - connected components diagnostic\n

    If a geneset/setname is passed to the function, the properties of the subgraph are evaluated

    '''

    logging.info("Evaluating network summary, please wait")
    network = rc.ReadTsv(network_file).get_network()

    if geneset_input_file:
        if not setname:
            logging.error('Missing setname name, specify  a unique setname')
        else:
            geneset = rc.ReadGmt(geneset_input_file).get_geneset(setname)
            for setname, item in geneset.items():
                graph = nx.subgraph(network, item)
                out.write_graph_summary(graph, text_output, setname + " on " + network_file)
                diagnostic.plot_connected_components(nx.connected_components(graph), c_components_figure_file)
                diagnostic.plot_degree(nx.degree(graph), degree_figure_file)

    else:
        out.write_graph_summary(network, text_output, network_file)
        diagnostic.plot_connected_components(nx.connected_components(network), c_components_figure_file)
        diagnostic.plot_degree(nx.degree(network), degree_figure_file)
    logging.info("Network summary completed")


################################################################################
######### SINGLE SET ANALYSES ##################################################
################################################################################


def test_topology_total_degree(
    network_file: "network file",
    geneset_file: "GMT geneset file",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    cores: "Number of cores for the multiprocessing" = 1,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    results_figure: "barplot of results, use pdf or png extension" = None,
    diagnostic_null_folder: "plot null distribution, pass the folder where all the figures are going to be saved "
                            "(one for each dataset)" = None):
    """
    Performs the analysis of total degree of the .

    It computes a p-value for the ratio of total degree of the geneset being bigger than the one expected by chance
    for a geneset of the same size.
    """
    logging.info("Evaluating the test topology total degree, please wait")
    network = rc.ReadTsv(network_file).get_network()
    data = rc.ReadTxt(geneset_file).get_data()
    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)

    setnames = [key for key in geneset.keys()]

    # Generate output
    output1 = out.Output(network_file, output_table, "topology_total_degree", geneset_file, setnames)
    logging.info("Results file = " + output1.output_table_results)

    # Create table
    output1.create_st_table_empirical()
    st_test = st.StatisticalTest(st.geneset_total_degree_statistic, network, degree_bins = n_bins)

    for setname, item in geneset.items():
        # Geneset smaller than size cut are not taken into consideration
        if len(item) > size_cut:
            item = set(item)
            observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item,
                                                                                     max_iter=number_of_permutations,
                                                                                     alternative="greater",
                                                                                     cores=cores)
            logging.info("Setname:" + setname)
            if n_mapped < size_cut:
                logging.info("%s removed from results since nodes mapped are < %d" % (setname, size_cut))
            else:
                logging.info("Observed: %g p-value: %g" % (observed, pvalue))
                # TODO Check line below
                logging.info("Null mean: %g null variance: %g".format(np.mean(null_d), np.var(null_d)))
                output1.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed,
                                                  pvalue, np.mean(null_d), np.var(null_d))
                if diagnostic_null_folder:
                    diagnostic.plot_null_distribution(null_d, observed, diagnostic_null_folder + setname +
                                                      '_total_degree_null_distribution.pdf', setname=setname)
    output1.close_temporary_table()
    if results_figure:
        paint.paint_datasets_stats(output1.output_table_results, results_figure, alternative='greater')
    logging.info("Test topology total degree completed")


def test_topology_internal_degree(
    network_file: "network file",
    geneset_file: "GMT geneset file",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    cores: "Number of cores for the multiprocessing" = 1,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    results_figure: "barplot of results, use pdf or png extension" = None,
    diagnostic_null_folder: "plot null distribution, pass the folder where all the figures are going to be saved "
                            "(one for each dataset)" = None,
    ):
    """
    Performs the analysis of internal degree.
    It computes a p-value for the ratio of internal degree of the geneset being bigger than the one expected by chance
    for a geneset of the same size.
    """

    network = rc.ReadTsv(network_file).get_network()
    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)
    setnames = [key for key in geneset.keys()]
    output1 = out.Output(network_file, output_table, "topology_internal_degree", geneset_file, setnames)
    logging.info("Results file = " + output1.output_table_results)
    output1.create_st_table_empirical()
    st_test = st.StatisticalTest(st.geneset_internal_degree_statistic, network, degree_bins = n_bins)
    for setname, item in geneset.items():
        item = set(item)
        if len(item) > size_cut:
            observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item,
                                                                                     max_iter=number_of_permutations,
                                                                                     alternative="greater", cores=cores)

            logging.info("Setname:" + setname)
            if n_mapped < size_cut:
                logging.info("%s remove from results since nodes mapped are < %d" % (setname, size_cut))
            else:
                logging.info("Observed: %g p-value: %g" % (observed, pvalue))
                output1.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed,
                                                  pvalue, np.mean(null_d), np.var(null_d))

                if diagnostic_null_folder:
                    diagnostic.plot_null_distribution(null_d, observed, diagnostic_null_folder + setname +
                                                      '_internal_degree_null_distribution.pdf', setname=setname)
    output1.close_temporary_table()
    if results_figure:
        paint.paint_datasets_stats(output1.output_table_results, results_figure, alternative='greater')


def test_topology_rwr(
    network_file: "network file, use a network with weights",
    geneset_file: "GMT geneset file",
    rwr_matrix_filename: "hdf5 RWR matrix obtained with pygna ",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    cores: "Number of cores for the multiprocessing" = 1,
    in_memory: "set if you want the large matrix to be read in memory" = False,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    results_figure: "barplot of results, use pdf or png extension" = None,
    diagnostic_null_folder: "plot null distribution, pass the folder where all the figures are going to be saved "
                            "(one for each dataset)" = None,
 ):
    """
    Performs the analysis of random walk probabilities.
    Given the RWR matrix, it compares the probability of walking between the genes in the geneset compared to
    those of walking between the nodes of a geneset with the same size
    """

    network = rc.ReadTsv(network_file).get_network()
    network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))
    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)
    rw_dict = {"nodes": read_distance_matrix(rwr_matrix_filename, in_memory=in_memory)[0],
               "matrix": read_distance_matrix(rwr_matrix_filename, in_memory=in_memory)[1]}

    setnames = [key for key in geneset.keys()]
    output1 = out.Output(network_file, output_table, "topology_rwr", geneset_file, setnames)

    logging.info("Results file = " + output1.output_table_results)
    output1.create_st_table_empirical()
    st_test = st.StatisticalTest(st.geneset_RW_statistic, network, rw_dict, degree_bins = n_bins)

    for setname, item in geneset.items():
        item = set(item)
        if len(item) > size_cut:
            # test
            observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item,
                                                                                     max_iter=number_of_permutations,
                                                                                     alternative="greater", cores=cores)
            logging.info("Setname:" + setname)
            if n_mapped < size_cut:
                logging.info("%s remove from results since nodes mapped are < %d" % (setname, size_cut))
            else:
                logging.info("Observed: %g p-value: %g" % (observed, pvalue))
                if diagnostic_null_folder:
                    diagnostic.plot_null_distribution(null_d, observed, diagnostic_null_folder + setname +
                                                      '_rwr_null_distribution.pdf', setname=setname)
                # saving output
                output1.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed,
                                                  pvalue, np.mean(null_d), np.var(null_d))
        else:
            logging.info("%s removed from results since nodes mapped are < %d" % (setname, size_cut))

    output1.close_temporary_table()
    if results_figure:
        paint.paint_datasets_stats(output1.output_table_results, results_figure, alternative='greater')


def test_topology_module(
    network_file: "network file",
    geneset_file: "GMT geneset file",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    cores: "Number of cores for the multiprocessing" = 1,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    output_lcc: "for creating a GMT file with the LCC lists pass a GMT filename" = None,
    results_figure: "barplot of results, use pdf or png extension" = None,
    diagnostic_null_folder: "plot null distribution, pass the folder where all the figures are going to be saved "
                            "(one for each dataset)" = None,
 ):
    """
    Performs geneset network topology module analysis.
    It computes a p-value for the largest connected component of the geneset being bigger than the one expected by chance
    for a geneset of the same size.
    """
    network = rc.ReadTsv(network_file).get_network()
    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)

    setnames = [key for key in geneset.keys()]
    output1 = out.Output(network_file, output_table, "topology_module", geneset_file, setnames)
    logging.info("Results file = " + output1.output_table_results)
    output1.create_st_table_empirical()

    st_test = st.StatisticalTest(st.geneset_module_statistic, network, degree_bins = n_bins)
    for setname, item in geneset.items():
        item = set(item)
        if len(item) > size_cut:
            if output_lcc:
                module = nx.subgraph(network, item)
                if len(module.nodes) > 0:
                    lcc = sorted(list(nx.connected_components(module)), key=len, reverse=True)[0]
                else:
                    lcc = []
                output1.add_GMT_entry(setname, "topology_module", lcc)

            observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item,
                                                                                     max_iter=number_of_permutations,
                                                                                     alternative="greater", cores=cores)
            logging.info("Setname:" + setname)
            if n_mapped < size_cut:
                logging.info("%s remove from results since nodes mapped are < %d" % (setname, size_cut))
            else:
                logging.info("Observed: %g p-value: %g" % (observed, pvalue))
                output1.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed,
                                                  pvalue, np.mean(null_d), np.var(null_d))
                if diagnostic_null_folder:
                    diagnostic.plot_null_distribution(null_d, observed, diagnostic_null_folder + setname +
                                                      '_module_null_distribution.pdf', setname=setname)
    output1.close_temporary_table()
    if output_lcc:
        output1.create_GMT_output(output_lcc)

    if results_figure:
        paint.paint_datasets_stats(output1.output_table_results, results_figure, alternative='greater')


def test_topology_sp(
    network_file: "network file",
    geneset_file: "GMT geneset file",
    distance_matrix_filename: "distance hdf5 matrix file generated by pygna",
    output_table: "output results table, use .csv extension",
    setname: "Geneset to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    cores: "Number of cores for the multiprocessing" = 1,
    in_memory: "set if you want the large matrix to be read in memory" = False,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    results_figure: "barplot of results, use pdf or png extension" = None,
    diagnostic_null_folder: "plot null distribution, pass the folder where all the figures are going to be saved "
                            "(one for each dataset)" = None,
                                                                ):

    """
    Performs geneset network topology shortest path analysis.

    It computes a p-value for the average shortest path length
    of the geneset being smaller than expected by chance
    for a geneset of the same size.
    """

    network = rc.ReadTsv(network_file).get_network()
    network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))

    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)

    diz = {"nodes": read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[0],
           "matrix": read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[1]}
    diz["matrix"] = diz["matrix"] + np.transpose(diz["matrix"])
    np.fill_diagonal(diz["matrix"], float("inf"))
    setnames = [key for key in geneset.keys()]

    output1 = out.Output(network_file, output_table, "topology_sp", geneset_file, setnames)
    logging.info("Results file = " + output1.output_table_results)
    output1.create_st_table_empirical()
    st_test = st.StatisticalTest(st.geneset_localisation_statistic, network, diz, degree_bins = n_bins)

    for setname, item in geneset.items():

        item = set(item)
        if len(item) > size_cut:
            logging.info("Setname:" + setname)
            observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(item, cores=cores,
                                                                                     max_iter=number_of_permutations)
            logging.info("Observed: %g p-value: %g" % (observed, pvalue))

            output1.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed, pvalue,
                                              np.mean(null_d), np.var(null_d))
            if diagnostic_null_folder:
                diagnostic.plot_null_distribution(null_d, observed, diagnostic_null_folder + setname +
                                                  '_sp_null_distribution.pdf', setname=setname, alternative="less")
        else:
            logging.info("%s remove from results since nodes mapped are < %d" % (setname, size_cut))
    output1.close_temporary_table()
    if results_figure:
        paint.paint_datasets_stats(output1.output_table_results, results_figure, alternative='less')


################################################################################
######### Diffusion test #######################################################
################################################################################
def test_diffusion_hotnet(network_file: "network file, use a network with weights",
                          geneset_file: "csv geneset file",
                          rwr_matrix_filename: "hdf5 RWR matrix obtained with pygna ",
                          output_table: "output results table, use .csv extension",
                          name_column: "Column to use as name (default is deseq2)" = "gene_name",
                          weight_column: "Column to use as weight (default is deseq2)" = "stat",
                          filter_column: "Column used to define the significant genes (default is deseq2)" = "padj",
                          filter_condition: "Condition for significance" = "less",
                          filter_threshold: "threshold for significance" = 0.01,
                          normalise: 'pass this flag for using only positive values in the analysis' = False,
                          size_cut: "removes all genesets with a mapped length < size_cut" = 20,
                          number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
                          cores: "Number of cores for the multiprocessing" = 1,
                          in_memory: "set if you want the large matrix to be read in memory" = False,
                          ):
    """
    Performs the analysis of random walk applying the weights of an upstream analysis.
    Given a csv file the user needs to specify the columns of interest and
    the threshold of significance.
    For the analysis the StatisticalDiffusion is used with hotnet_diffusion_statistic
    function.
    """

    # Reading network file
    network = rc.ReadTsv(network_file).get_network()
    network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))

    # Read geneset
    table = rc.ReadCsv(geneset_file, column_to_fill=name_column).get_data()
    if len(table.columns) < 2:
        logging.error("Error: the function takes a csv file as input, the read file has less than 2 columns, "
                      "check that the table is comma separated")

    # Filter table for significant genes
    table[name_column] = table[name_column].fillna(0).apply(str)
    table = pe.TableElaboration.clean_table(table=table, stat_col=weight_column)
    geneset = utils.filter_table(table, filter_column=filter_column, alternative=filter_condition,
                                 threshold=filter_threshold)[name_column]
    if normalise:
        table[weight_column] = np.abs(table[weight_column].values)

    if len(geneset) < size_cut:
        logging.error('The number of significant genes is lower than %d. \
                    \n Change size_cut if necessary' % size_cut)

    # Read RWR matrix
    rw_dict = {"nodes": read_distance_matrix(rwr_matrix_filename, in_memory=in_memory)[0],
               "matrix": read_distance_matrix(rwr_matrix_filename, in_memory=in_memory)[1]}
    # setting output
    output1 = out.Output(network_file, output_table, "diffusion", geneset_file, geneset_file)
    output1.create_st_table_empirical()
    logging.info("Results file = " + output1.output_table_results)

    # initialising test
    st_test = sd.DiffusionTest(sd.hotnet_diffusion_statistic, rw_dict["nodes"], rw_dict["matrix"], table,
                               names_col=name_column, weights_col=weight_column)

    observed, pvalue, null_d, n_mapped, n_geneset = st_test.empirical_pvalue(geneset, max_iter=number_of_permutations,
                                                                             alternative="greater", cores=cores)
    if n_mapped < size_cut:
        logging.info("Results removed, since nodes mapped are < %d" % size_cut)
    else:
        logging.info("Observed: %g p-value: %g" % (observed, pvalue))
        output1.update_st_table_empirical(geneset_file, n_mapped, n_geneset, number_of_permutations, observed, pvalue,
                                          np.mean(null_d), np.var(null_d))
    output1.close_temporary_table()

    # if results_figure:
    #    paint.paint_diffusion_matrix(output1.output_table_results, results_figure, alternative='greater',)


################################################################################
######### ASSOCIATIONS and COMPARISONS #########################################
################################################################################

def test_association_sp(
    network_file: "network file",
    file_geneset_a: "GMT geneset file, if it's the only parameter passed the analysis is gonna be run on all the "
                    "pair of datasets, otherwise specify the other files and setnames",
    distance_matrix_filename: "distance matrix file generated by pygna",
    output_table: "output results table, use .csv extension",
    setname_a: "Geneset A to analyse" = None,
    file_geneset_b: "GMT geneset file" = None,
    setname_b: "Geneset B to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    keep: "if true, keeps the geneset B not permuted" = False,
    cores: "Number of cores for the multiprocessing" = 1,
    in_memory: "set if you want the large matrix to be read in memory" = False,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    results_figure: "barplot of results, use pdf or png extension" = None,
):
    """
    Performs comparison of network location analysis. If the flag
    --keep  is passed, the B geneset is kept fixed, and doesnt't get permuted.

    It computes a p-value for the shortest path distance between two genesets being smaller than expected by chance
    If only A_geneset_file is passed the analysis is run on all the pair of sets in the file, if both
    A_geneset_file and B_geneset_file are passed, one can specify the setnames for both, if there is only one
    geneset in the file, setname_X can be omitted, if both sets are in the same file, B_geneset_file can be not
    specified, but setnames are needed.
    """

    if keep:
        analysis_name_str = "association_sp"
    else:
        analysis_name_str = "comparison_sp"

    network = rc.ReadTsv(network_file).get_network()
    network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))

    # Read matrix
    sp_diz = {"nodes": read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[0],
              "matrix": read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[1]}

    sp_diz["matrix"] = sp_diz["matrix"] + np.transpose(sp_diz["matrix"])
    np.fill_diagonal(sp_diz["matrix"], np.inf)

    # Managing the different genesets
    if setname_a and setname_b is None and file_geneset_b is None:
        logging.error(" this analysis requires at least two genesets ")

    geneset_a = rc.ReadGmt(file_geneset_a).get_geneset(setname_a)
    if file_geneset_b:
        geneset_b = rc.ReadGmt(file_geneset_b).get_geneset(setname_b)
    else:
        if setname_b:
            geneset_b = rc.ReadGmt(file_geneset_a).get_geneset(setname_b)
        else:
            geneset_b = None

    st_comparison = sc.StatisticalComparison(sc.comparison_shortest_path, network, diz=sp_diz, n_proc=cores, degree_bins = n_bins)

    if not geneset_b:  # Analysis of genesets inside a single file
        logging.info("Analysing all the sets in " + file_geneset_a)
        setnames = [key for key in geneset_a.keys()]

        # Creating the output table
        output1 = out.Output(network_file, output_table, analysis_name_str, file_geneset_a, setnames)
        logging.info("Results file = " + output1.output_table_results)
        output1.create_comparison_table_empirical()

        for pair in itertools.combinations(setnames, 2):
            if len(set(geneset_a[pair[0]])) > size_cut and len(set(geneset_a[pair[1]])) > size_cut:
                logging.info("Analysing " + str(pair[0]) + " and " + str(pair[1]))

                n_overlaps = len(set(geneset_a[pair[0]]).intersection(set(geneset_a[pair[1]])))
                observed, pvalue, null_d, a_mapped, b_mapped = st_comparison.comparison_empirical_pvalue(
                    set(geneset_a[pair[0]]), set(geneset_a[pair[1]]), max_iter=number_of_permutations, keep=keep)
                # Save the results

                output1.update_comparison_table_empirical(pair[0], pair[1], len(set(geneset_a[pair[0]])), a_mapped,
                                                          len(set(geneset_a[pair[1]])), b_mapped, n_overlaps,
                                                          number_of_permutations, observed, pvalue, np.mean(null_d),
                                                          np.var(null_d))
            else:
                logging.warning("Geneset A has %d terms and Geneset B has %d terms. \
                \nOne of them is too short, analysis not done" % (len(set(geneset_a[pair[0]])),
                                                                  len(set(geneset_a[pair[1]]))))

    else:  # Analysis of genesets into two different GMT files

        logging.info("geneset_a contains %d sets", (len(geneset_a)))
        sets_a = [key for key in geneset_a.keys()]
        logging.info("geneset_b contains %d sets", (len(geneset_b)))
        sets_b = [key for key in geneset_b.keys()]
        output1 = out.Output(network_file, output_table, analysis_name_str, file_geneset_a, sets_a, file_geneset_b,
                             sets_b)
        logging.info("Results file = " + output1.output_table_results)
        output1.create_comparison_table_empirical()
        for set_A, item_A in geneset_a.items():
            for set_B, item_B in geneset_b.items():
                n_overlaps = len(set(item_A).intersection(set(item_B)))
                if len(item_A) > size_cut and len(item_B) > size_cut:
                    observed, pvalue, null_d, a_mapped, b_mapped = st_comparison.comparison_empirical_pvalue(
                        set(item_A), set(item_B), max_iter=number_of_permutations, keep=keep)

                    logging.info("Observed: %g p-value: %g" % (observed, pvalue))
                    output1.update_comparison_table_empirical(set_A, set_B, len(set(item_A)), a_mapped,
                                                              len(set(item_B)), b_mapped, n_overlaps,
                                                              number_of_permutations, observed, pvalue, np.mean(null_d),
                                                              np.var(null_d))
    output1.close_temporary_table()
    if results_figure:
        paint.paint_comparison_matrix(output1.output_table_results, results_figure)


def test_association_rwr(
    network_file: "network file",
    file_geneset_a: "GMT geneset file",
    rwr_matrix_filename: ".hdf5 file with the RWR matrix obtained by pygna",
    output_table: "output results table, use .csv extension",
    setname_a: "Geneset A to analyse" = None,
    file_geneset_b: "GMT geneset file" = None,
    setname_b: "Geneset B to analyse" = None,
    size_cut: "removes all genesets with a mapped length < size_cut" = 20,
    keep: "if true, keeps the geneset B unpermuted" = False,
    cores: "Number of cores for the multiprocessing" = 1,
    in_memory: "set if you want the large matrix to be read in memory" = False,
    number_of_permutations: "number of permutations for computing the empirical pvalue" = 500,
    n_bins: 'if >1 applies degree correction by binning the node degrees and sampling according to geneset distribution' = 1,
    results_figure: "heatmap of results" = None,
):
    """
    Performs comparison of network location analysis.

    It computes a p-value for the shortest path distance
    between two genesets being smaller than expected by chance.

    If only A_geneset_file is passed the analysis is run on all the pair of sets in the file, if both
    A_geneset_file and B_geneset_file are passed, one can specify the setnames for both, if there is only one
    geneset in the file, setname_X can be omitted, if both sets are in the same file, B_geneset_file can be not
    specified, but setnames are needed.
    """

    if keep:
        analysis_name_str = "association_rwr"
    else:
        analysis_name_str = "comparison_rwr"

    network = rc.ReadTsv(network_file).get_network()
    network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))

    # Read datasets
    if setname_a and setname_b is None and file_geneset_b is None:
        logging.error(" this analysis requires at least two genesets ")

    rw_dict = {"nodes": read_distance_matrix(rwr_matrix_filename, in_memory=in_memory)[0],
               "matrix": read_distance_matrix(rwr_matrix_filename, in_memory=in_memory)[1]}

    geneset_a = rc.ReadGmt(file_geneset_a).get_geneset(setname_a)
    if file_geneset_b:
        geneset_b = rc.ReadGmt(file_geneset_b).get_geneset(setname_b)
    else:
        if setname_b:
            geneset_b = rc.ReadGmt(file_geneset_a).get_geneset(setname_b)
        else:
            geneset_b = None

    st_comparison = sc.StatisticalComparison(sc.comparison_random_walk, network, n_proc=cores, diz=rw_dict, degree_bins = n_bins)
    if not geneset_b:
        logging.info("Analysing all the sets in " + file_geneset_a)
        setnames = [key for key in geneset_a.keys()]
        output1 = out.Output(network_file, output_table, analysis_name_str, file_geneset_a, setnames)
        logging.info("Results file = " + output1.output_table_results)
        output1.create_comparison_table_empirical()

        for pair in itertools.combinations(setnames, 2):
            if len(set(geneset_a[pair[0]])) > size_cut and len(set(geneset_a[pair[1]])) > size_cut:
                logging.info("Analysing " + str(pair[0]) + " and " + str(pair[1]))
                n_overlaps = len(set(geneset_a[pair[0]]).intersection(set(geneset_a[pair[1]])))
                observed, pvalue, null_d, a_mapped, b_mapped = st_comparison.comparison_empirical_pvalue(
                    set(geneset_a[pair[0]]), set(geneset_a[pair[1]]), max_iter=number_of_permutations,
                    alternative="greater", keep=keep)

                output1.update_comparison_table_empirical(pair[0], pair[1], len(set(geneset_a[pair[0]])), a_mapped,
                                                          len(set(geneset_a[pair[1]])), b_mapped, n_overlaps,
                                                          number_of_permutations, observed, pvalue, np.mean(null_d),
                                                          np.var(null_d))
            else:
                logging.warning("Geneset A has %d terms and Geneset B has %d terms. \
                \nOne of them is too short, analysis not done" % (len(set(geneset_a[pair[0]])),
                                                                  len(set(geneset_a[pair[1]]))))

    else:
        logging.info("geneset_a contains %d sets" % (len(geneset_a)))
        sets_a = [key for key in geneset_a.keys()]
        logging.info("Setnames in A: " + str(sets_a))
        logging.info("geneset_b contains %d sets" % (len(geneset_b)))
        sets_b = [key for key in geneset_b.keys()]
        logging.info("Setnames in B: " + str(sets_b))
        output1 = out.Output(network_file, output_table, analysis_name_str, file_geneset_a, sets_a, file_geneset_b,
                             sets_b)
        logging.info("Results file = " + output1.output_table_results)
        output1.create_comparison_table_empirical()

        for set_A, item_A in geneset_a.items():
            for set_B, item_B in geneset_b.items():

                if len(item_A) > size_cut and len(item_B) > size_cut:
                    logging.info("Analysing " + str(set_A) + " and " + str(set_B))
                    n_overlaps = len(set(item_A).intersection(set(item_B)))
                    observed, pvalue, null_d, a_mapped, b_mapped = st_comparison.comparison_empirical_pvalue(
                        set(item_A), set(item_B), max_iter=number_of_permutations, alternative="greater", keep=keep)
                    logging.info("Observed: %g p-value: %g" % (observed, pvalue))

                    output1.update_comparison_table_empirical(set_A, set_B, len(set(item_A)), a_mapped,
                                                              len(set(item_B)), b_mapped, n_overlaps,
                                                              number_of_permutations, observed, pvalue, np.mean(null_d),
                                                              np.var(null_d))
                else:
                    logging.warning("Geneset A has %d terms and Geneset B has %d terms. \
                    \nOne of them is too short, analysis not done" % (len(set(item_A)), len(set(item_B))))

    output1.close_temporary_table()
    if results_figure:
        paint.paint_comparison_matrix(output1.output_table_results, results_figure, rwr=True)


################################################################################
######### BUILDING FUNCTIONS ###################################################
################################################################################

def build_distance_matrix(
    network_file: "network file",
    output_file: "distance matrix output file, use .hdf5",
    giant_component_only: "compute the shortest paths only for nodes in the giant component" = True,
):
    """
    Build a shortest path distance matrix for a given network.
    Matrix can be saved as a .txt file or a .hdf5 one.
    """
    logging.info("Converting distance matrix, please wait...")
    network = rc.ReadTsv(network_file).get_network()
    if giant_component_only:
        network = network.subgraph(max(nx.connected_components(network), key=len))

    distance_matrix = nx.all_pairs_shortest_path_length(network)

    nodes = list(network.nodes())

    if output_file.endswith(".hdf5"):
        with tables.open_file(output_file, mode="w") as hdf5_file:

            # create a hdf5 file with two objects:
            # - one is the nodes array,
            hdf5_nodes = hdf5_file.create_array(hdf5_file.root, "nodes", nodes)
            # -  the other is the shortest path distance matrix
            hdf5_data = hdf5_file.create_array(hdf5_file.root, "matrix", np.zeros((len(nodes), len(nodes))))
            for node_i, k in distance_matrix:
                i = nodes.index(node_i)
                for node_j, sp in k.items():
                    j = nodes.index(node_j)
                    if j >= i:  # saves only the upper triangular matrix
                        hdf5_data[i, j] = sp
        hdf5_file.close()
    else:
        logging.error("Pass an hd5f file")


def build_rwr_diffusion(
    network_file: "network file",
    beta=0.85,
    output_file: "distance matrix output file (use .hdf5) " = None,
):
    """
    Build the RWR_diffusion_matrix
    """

    network = rc.ReadTsv(network_file).get_network()
    network = network.subgraph(max(nx.connected_components(network), key=len))
    nodes = list(network.nodes())

    logging.info("Beginning to calculate RWR matrix")

    a = nx.adjacency_matrix(network)
    k = 1 / np.array(list(dict(network.degree()).values()))
    d = scipy.sparse.dia_matrix((k, [0]), shape=a.shape)
    a = a.dot(d)
    n = np.shape(a)[1]

    if output_file.endswith(".hdf5"):
        with tables.open_file(output_file, mode="w") as hdf5_file:
            # create a hdf5 file with two objects:
            # - one is the nodes array,
            hdf5_file.create_array(hdf5_file.root, "nodes", nodes)
            # -  the other is the RWR matrix
            hdf5_file.create_array(hdf5_file.root, "matrix", beta * np.linalg.inv(np.eye(n) - (1.0 - beta) * a))
            logging.info("Saving network")
            hdf5_file.close()
    else:
        return beta * np.linalg.inv(np.eye(n) - (1.0 - beta) * a)


def network_graphml(
    network_file: "network file",
    geneset_file: "geneset file",
    output_file: "graphml file for network for visualisation",
    setname: "set name" = None,
    giant_component_only: "saves only the giant component of the network" = True,
    minimal: 'saves only the minimal graph' = False,
):
    """
    This function generates a graphml file with nodes annotation.
    Given a geneset, with k setnames, each node has k False/True
    annotations for each set.

    Warning: without minimal, this function saves the full network.
    The minimal graph saves only the nodes in the geneset and those that
    connect them with a shortest path.
    """

    network = rc.ReadTsv(network_file).get_network()
    geneset = rc.ReadGmt(geneset_file).get_geneset(setname)

    if giant_component_only:
        network = network.subgraph(max(nx.connected_components(network), key=len))

    dict_nodes = {}
    if minimal:
        for setname in geneset:
            new_network_minimal = nx.Graph()
            for s in geneset[setname]:
                if s in network.nodes():
                    l0 = np.inf
                    for t in geneset[setname]:
                        if (t in network.nodes()) & (s != t):
                            path = nx.shortest_path(network, source=s, target=t)
                            if l0 > len(path):
                                l0 = len(path)
                                new_network_minimal.add_path(path)
            for n in new_network_minimal.nodes():
                if n in geneset[setname]:
                    dict_nodes[n] = True
                else:
                    dict_nodes[n] = False
            nx.set_node_attributes(new_network_minimal, dict_nodes, setname)

    else:
        for setname in geneset:
            for n in network.nodes():
                if n in geneset[setname]:
                    dict_nodes[n] = True
                else:
                    dict_nodes[n] = False
            nx.set_node_attributes(network, dict_nodes, setname)

    nx.write_graphml(network, output_file)


def get_connected_components(network_file: "network tsv file",
                             output_gmt: "The output file name (should be gmt)",
                             name : 'pass a name for the putput gmt terms',
                             geneset_file: "GMT of the geneset file, is a file is passed please add the setname" = None,
                             setname: "The setname to analyse" = None,
                             graphml: "Pass a graphml filename to show the results on Cytoscape" = None,
                             threshold: 'ignores all CC smaller than this value' = 1,
                             convert_entrez: "pass flag to convert EntrezID->Symbol" = False):
    """
    This function evaluate all the connected components in the subgraph pf the network with a given setname.
    Multiple setnames can be passed to this function to analyze all of them in a run.
    The file produces a GMT output and optionally a plot of the subnetwork with the connected components analysed.
    Please notice that to convert the entrezID into Symbols, a stable internet connection is required
    """

    network = rc.ReadTsv(network_file).get_network()

    if type(geneset_file)==str:
        if setname==None:
            geneset = rc.ReadGmt(geneset_file).get_geneset()
            logging.error('using only first entry of the gmt: %s' %(list(geneset.keys())[0]))
            geneset = geneset[list(geneset.keys())[0]]
        else:
            geneset = rc.ReadGmt(geneset_file).get_geneset(setname)[setname]
        network = nx.subgraph(network, list(set(geneset)))

    print(network)

    output1 = out.Output(network_file, 'o.csv', "network_gmt", geneset_file, setname)
    output1.create_st_table_empirical()
    cclist = list()

    mg = mygene.MyGeneInfo()

    connected_components = nx.connected_components(network)
    i = 0
    for cc in connected_components:
        print(cc)
        if len(cc) > threshold:
            i = i + 1
            if convert_entrez:
                cc = mg.querymany(list(cc), scopes='entrezgene', fields='symbol', species='human')
                gene_list = list()
                [gene_list.append(e["symbol"]) for e in cc]
                cc = gene_list
            cclist.append(cc)

            nodes = {}
            for node in cc:
                nodes[node] = i
            output1.add_GMT_entry(name + "_" + str(i), "connected components", cc)

    output1.create_GMT_output(output_gmt)

    if len(network.nodes())>1000:
        logging.info('There are more than 100 nodes in the network, the graphml file might be very large.')

    nx.write_graphml(network, graphml)




import logging
import argh
import pygna.command as cmd
import pygna.painter as paint
import pygna.utils as utils
import pygna.block_model as bm
import pygna.degree_model as dm

"""
autodoc
"""

logging.basicConfig(level=logging.INFO)


def main():
    argh.dispatch_commands([
        # network summary and graph file
        cmd.network_summary,
        cmd.network_graphml,
        cmd.get_connected_components,
        # geneset network topology analyses
        cmd.test_topology_total_degree,
        cmd.test_topology_internal_degree,
        cmd.test_topology_module,
        cmd.test_topology_sp,
        cmd.test_topology_rwr,
        cmd.test_diffusion_hotnet,
        # comparison analysis
        cmd.test_association_sp,
        cmd.test_association_rwr,
        # building functions
        cmd.build_distance_matrix,
        cmd.build_rwr_diffusion,
        # paint
        paint.paint_datasets_stats,
        paint.paint_comparison_matrix,
        paint.plot_adjacency,
        paint.paint_volcano_plot,

        paint.paint_summary_gnt,
        # utils
        utils.convert_gmt,
        utils.geneset_from_table,
        utils.convert_csv,
        utils.generate_group_gmt,
        # simulations
        bm.generate_gnt_sbm,
        bm.generate_gna_sbm,
        dm.generate_hdn_network,
        bm.generate_sbm_network,
        bm.generate_sbm2_network,

        dm.hdn_add_partial,
        dm.hdn_add_extended,
        dm.hdn_add_branching,

         ], )

if __name__ == "__main__":
    """
    MAIN
    """
    main()



import logging
import networkx as nx
import os
import numpy as np
import statsmodels.stats.multitest as multi
import pandas as pd
import scipy
import tempfile
import shutil


class Output:
    """
    This class is used to print different data on files
    """

    def __init__(self, network_filename: str, output_table_results_file: str, analysis: str, geneset_file: str,
                 setnames: list, geneset_file_B: str = None, setnames_B: list = None):
        """
        :param network_filename: the file containing the network
        :param output_table_results_file: the output table that contains the results to use
        :param analysis: the type of analysis performed
        :param geneset_file: the geneset file use
        :param setnames: the names of the first geneset
        :param geneset_file_B: the second geneset file to use
        :param setnames_B: the names of the second geneset
        """
        self.network_filename = network_filename
        self.analysis = analysis
        self.output_table_results = output_table_results_file
        self.output_gmt = None
        self.text = []
        self.geneset_filename = geneset_file
        self.setnames = setnames
        self.geneset_filename_B = geneset_file_B
        self.setnames_B = setnames_B
        self.diffusion_matrix_file = None
        self.GMT_dict = {}

        if not self.output_table_results.endswith('.csv'):
            logging.warning('The output table is saved as csv file, the name does not match the file extension')

    def set_diffusion_matrix(self, diffusion_matrix_file: str) -> None:
        """
        Set the diffusion matrix file

        :param diffusion_matrix_file: set the diffusion matrix file to use

        Example
        ______
        >>> setnames = ["A", "B", "C"]
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> out.set_diffusion_matrix("diffusion_matrix.csv")
        """
        self.diffusion_matrix_file = diffusion_matrix_file

    # Tables for stats
    def create_st_table_empirical(self) -> None:
        """
        Create the headings of the table that are going to be wrinnte in the csv file

        Example
        _______
        >>> setnames = ["A", "B", "C"]
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> out.create_st_table_empirical()
        """
        tmp = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
        self.table_file_name = tmp.name
        try:
            tmp.write("analysis,setname,n_mapped,n_geneset,number_of_permutations,observed,empirical_pvalue,mean(null),"
                      "var(null),network,geneset\n")
        finally:
            tmp.close()

    def close_temporary_table(self) -> None:
        """
        Remove the temporary file

        Example
        _______
        >>> setnames = ["A", "B", "C"]
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> out.create_st_table_empirical()
        >>> out.close_temporary_table()
        """
        shutil.copy(self.table_file_name, self.output_table_results)
        os.remove(self.table_file_name)

    def update_st_table_empirical(self, setname: str, n_mapped: int, n_geneset: int, number_of_permutations: int,
                                  observed: int, empirical_pvalue: float, mean_null: np.mean, var_null: np.var) -> None:
        """
        Update the table content, adding a new line to the file

        :param setname: the name of the geneset
        :param n_mapped: the number of mapped genes
        :param n_geneset: the number of genesets
        :param number_of_permutations: the number of permutations
        :param observed: value of observed genes
        :param empirical_pvalue: value of the empirical p-value
        :param mean_null: mean of the null distribution
        :param var_null: var of the null distribution

        Example
        _______
        >>> setnames = ["A", "B", "C"]
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> out.create_st_table_empirical()
        >>> out.update_st_table_empirical(setname, n_mapped, n_geneset, number_of_permutations, observed, pvalue=0.001, mean_null=np.mean(0.11), var_null=np.var(0.2))
        """
        setname = setname.replace(",", "_")
        with open(self.table_file_name, "a") as f:
            f.write(",".join([str(x) for x in
                              [self.analysis, setname, n_mapped, n_geneset, number_of_permutations, observed,
                               empirical_pvalue, mean_null, var_null, self.network_filename,
                               self.geneset_filename]]) + "\n")

    # Tables for comparisons
    def create_comparison_table_empirical(self) -> None:
        """
        Write the hadings for the comparison table

        Example
        _______
        >>> setnames = ["A", "B", "C"]
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> out.create_comparison_table_empirical()
        """
        tmp = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
        self.table_file_name = tmp.name
        try:
            tmp.write("analysis,setname_A,setname_B,n_geneset_A,n_mapped_A,n_geneset_B,n_mapped_B,n_overlaps,"
                      "number_of_permutations,observed,empirical_pvalue,mean(null),var(null),network\n")
        finally:
            tmp.close()

    def update_comparison_table_empirical(self, setname_A: str, setname_B: str, n_geneset_A: int, n_mapped_A: int,
                                          n_geneset_B: int, n_mapped_B: int, n_overlaps: int,
                                          number_of_permutations: int, observed: int, empirical_pvalue: float,
                                          mean_null: np.mean, var_null: np.var) -> None:
        """
        Update the content of the comparison table, adding a new row on the file

        :param setname_A: the name of the geneset A
        :param setname_B: the name of the geneset B
        :param n_geneset_A: the number of genes in the geneset A
        :param n_mapped_A: the number of mapped genes in geneset A
        :param n_geneset_B: the number of genes in the geneset B
        :param n_mapped_B: the number of mapped genes in geneset B
        :param n_overlaps: the number of overlaps
        :param number_of_permutations: number of performed permutations
        :param observed: number of observed genes
        :param empirical_pvalue: value of the empirical pvalue
        :param mean_null: mean of the null distribution
        :param var_null: variance of the null distribution

        Example
        _______
        >>> import itertools
        >>> import pygna.command as cm
        >>> import pygna.reading_class as rc
        >>> import pygna.statistical_comparison as sc
        >>> geneset_a = rc.ReadGmt("genset_file").get_geneset("brca")
        >>> setnames = [key for key in geneset_a.keys()]
        >>> network = rc.ReadTsv("network_file.tsv").get_network()
        >>> distance_matrix_filename = "distance_matrix.tsv"
        >>> in_memory = True
        >>> network = nx.Graph(network.subgraph(max(nx.connected_components(network), key=len)))
        >>> sp_diz = {"nodes": cm.read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[0],
        ...           "matrix": cm.read_distance_matrix(distance_matrix_filename, in_memory=in_memory)[1]}
        >>> st_comparison = sc.StatisticalComparison(sc.comparison_shortest_path, network, diz=sp_diz, n_proc=2)
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> for pair in itertools.combinations(setnames, 2):
        ...     observed, pvalue, null_d, a_mapped, b_mapped = st_comparison.comparison_empirical_pvalue(set(geneset_a[pair[0]]), set(geneset_a[pair[1]]), max_iter=number_of_permutations)
        ...     out.update_comparison_table_empirical(pair[0], pair[1], len(set(geneset_a[pair[0]])), a_mapped, len(set(geneset_a[pair[1]])), b_mapped, n_overlaps, number_of_permutations, observed, pvalue, np.mean(null_d), np.var(null_d))
        """
        setname_A = setname_A.replace(",", "_")
        setname_B = setname_B.replace(",", "_")
        with open(self.table_file_name, "a") as f:
            f.write(",".join([str(x) for x in
                              [self.analysis, setname_A, setname_B, n_geneset_A, n_mapped_A, n_geneset_B, n_mapped_B,
                               n_overlaps, number_of_permutations, observed, empirical_pvalue, mean_null, var_null,
                               self.network_filename]]) + "\n")

    def add_GMT_entry(self, key: str, descriptor: str, gene_list: str) -> None:
        """
        Add a gmt entry in the GMT file

        :param key: the key name to store
        :param descriptor: the descriptor of the gene list
        :param gene_list: the gene list to write

        Example
        _______
        >>> geneset = rc.ReadGmt("geneset_file.csv").get_geneset("brca")
        >>> setnames = [key for key in geneset.keys()]
        >>> import pygna.reading_class as rc
        >>> network = rc.ReadTsv("network_file.tsv").get_network()
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> for setname, item in geneset.items():
        ...     item = set(item)
        ...     module = nx.subgraph(network, item)
        ...     lcc = sorted(list(nx.connected_components(module)), key=len, reverse=True)[0]
        ...     out.add_GMT_entry("brca", "topology_module", lcc)
        """
        try:
            self.GMT_dict[key]
        except KeyError:
            self.GMT_dict[key] = {}
            self.GMT_dict[key]["descriptor"] = descriptor
            self.GMT_dict[key]["genes"] = gene_list

        else:
            logging.warning("Key Already Exists: " + str(key))

    def create_GMT_output(self, output_gmt: str) -> None:
        """
        Write the GMT line on the GMT file

        :param output_gmt: the GMT to print

        Example
        _______
        >>> geneset = rc.ReadGmt("geneset_file.csv").get_geneset("brca")
        >>> setnames = [key for key in geneset.keys()]
        >>> import pygna.reading_class as rc
        >>> network = rc.ReadTsv("network_file.tsv").get_network()
        >>> out = Output("networkfile.tsv", "results.csv", "myanalysis", "genset_a.csv", setnames)
        >>> for setname, item in geneset.items():
        ...     item = set(item)
        ...     module = nx.subgraph(network, item)
        ...     lcc = sorted(list(nx.connected_components(module)), key=len, reverse=True)[0]
        ...     out.add_GMT_entry("brca", "topology_module", lcc)
        >>> out.create_GMT_output("output_lcc.gmt")
        """
        self.output_gmt = output_gmt
        print_GMT(self.GMT_dict, self.output_gmt)


def print_GMT(gmt_dictionary: dict, output_file: str) -> None:
    """
    Save the dictionary on a GMT file

    :param gmt_dictionary: the dictionary containing the data
    :param output_file: the file to save the data

    Example
    _______
    >>> gmt_dict = {"key": "dict_sets"}
    >>> print_GMT(gmt_dict, "mygmt.gmt")
    """
    with open(output_file, "w") as f:
        f.write("")

    for key, dict_set in gmt_dictionary.items():
        with open(output_file, "a") as f:
            genes_dict = '\t'.join(map(str, dict_set["genes"]))
            f.write(str(key) + "\t" + str(dict_set["descriptor"]) + "\t" + genes_dict + "\n")


def apply_multiple_testing_correction(table_file: str, pval_col: str = "empirical_pvalue", method: str = "fdr_bh",
                                      threshold: float = 0.1) -> None:
    """
    Apply the multiple testing correction and save the file in a csv file

    :param table_file: the name of the file to read
    :param pval_col: the name column containing the empirical pvalue
    :param method: the correction method to use
    :param threshold: the threshold to use in the method

    Example
    _______
    >>> table_filename = "pygna_comparison_results.csv"
    >>> apply_multiple_testing_correction(table_filename, pval_col="empirical_pvalue", method="fdr_bh", threshold=0.1)
    """
    with open(table_file, "r+") as f:
        table = pd.read_csv(f)

    rejects, pval, k, bonf = multi.multipletests(table[pval_col].values, alpha=float(threshold), method=method)
    table["rejects"] = rejects
    table["bh_pvalue"] = pval
    table["k"] = k
    table["bonf"] = bonf

    table = table.sort_values(by="bh_pvalue")

    table.to_csv(table_file, index=False)


def write_graph_summary(graph: nx.Graph, output_file: str, net_name: str = None) -> None:
    """
    This function takes a graph as input and writes the network properties in a text file

    :param graph: the graph to print
    :param output_file: the name of the file to print
    :param net_name: the name of the network

    Example
    _______
    >>> import pygna.reading_class as rc
    >>> text_output = "My summary stats"
    >>> network = rc.ReadTsv("mynetwork.tsv").get_network()
    >>> write_graph_summary(network, text_output, "mynetwork.tsv")
    """

    if not net_name:
        net_name = 'network'

    D = dict(nx.degree(graph))
    degree = np.array(list(dict(nx.degree(graph)).values()))

    n_nodes = nx.number_of_nodes(graph)
    n_edges = nx.number_of_edges(graph)

    degrees = {k: v for k, v in D.items()}
    degrees = sorted(degrees.items(), key=lambda kv: kv[1])

    density = (2 * n_edges) / ((n_nodes) * (n_nodes - 1))

    with open(output_file, "w") as file1:
        file1.write("Network Summary for %s " % str(net_name))
        file1.write("\n---------------------------------------------------\n")
        file1.write("\nInfo: " + nx.info(graph))
        file1.write("\nOther Properties::\n ")
        file1.write("\n\t- Density: " + str(density))
        file1.write("\n\t- min degree = " + str(np.min(degree)))
        file1.write("\n\t- max degree = " + str(np.max(degree)))
        file1.write("\n\t- median degree = " + str(np.median(degree)))
        file1.write("\n\t- degree mode = " + str(scipy.stats.mode(degree)))
        file1.write("\n\t- disconnected nodes = " + str(np.sum(degree == 0)))
        file1.write("\n\t- average clustering" + str(nx.average_clustering(graph)))

    largest_cc = nx.Graph(graph.subgraph(max(nx.connected_components(graph), key=len)))

    D = dict(nx.degree(largest_cc))
    degree = np.array(list(dict(nx.degree(largest_cc)).values()))

    n_nodes = nx.number_of_nodes(largest_cc)
    n_edges = nx.number_of_edges(largest_cc)

    degrees = {k: v for k, v in D.items()}
    degrees = sorted(degrees.items(), key=lambda kv: kv[1])

    density = (2 * n_edges) / ((n_nodes) * (n_nodes - 1))

    with open(output_file, "a") as file1:
        file1.write("\n\nLargest Connected Component Summary for %s \n " % str(net_name))
        file1.write(
            "-----------------------------------------------------------------\n"
        )
        file1.write("\nInfo: " + nx.info(largest_cc))
        file1.write("\nLargest Connected Component::\n")

        file1.write("\n\t- Density: " + str(density))
        file1.write("\n\t- min degree = " + str(np.min(degree)))
        file1.write("\n\t- max degree = " + str(np.max(degree)))
        file1.write("\n\t- median degree = " + str(np.median(degree)))
        file1.write("\n\t- degree mode = " + str(scipy.stats.mode(degree)))
        file1.write("\n\t- disconnected nodes = " + str(np.sum(degree == 0)))
        file1.write("\n\t- average clustering" + str(nx.average_clustering(largest_cc)))



import logging
import numpy as np
import pandas as pd
import multiprocessing


def map_table2matrix(node_list, x):
    return node_list.index(x)


class DiffusionTest:
    """
    This class elaborates the diffusion statistics. It elaborates the diffusion tests over the given network.
    Please refer to the single method documentation for the returning values
    """

    def __init__(self, test_statistic, nodes: list, diffusion_matrix: np.matrix, weights_table: pd.DataFrame,
                 names_col: str = "name", weights_col: str = "stat", diz: dict = {}):
        """
        :param test_statistic: the function to be applied to calculate the statistics
        :param nodes: the network nodes
        :param diffusion_matrix: the diffusion matrix to apply the test statistic
        :param weights_table: the table with the genes
        :param names_col: the column name to read the elements
        :param weights_col: the column name to read the weights
        :param diz: the dictionary used in the observation
        """

        self.test_statistic = test_statistic
        self.nodes = nodes
        self.diffusion_matrix = diffusion_matrix
        self.diz = diz
        self.universe = set(self.nodes)

        self.table = weights_table[weights_table[names_col].isin(self.nodes)]
        logging.info("weights table: %d rows" % len(weights_table))
        logging.info("mapped table: %d mapped rows" % len(self.table))

        if len(self.table) < 10:
            logging.warning("there are less than 10 elements in the table that are mapped to the universe")

        self.table_index = [map_table2matrix(self.nodes, i) for i in self.table[names_col].values.tolist()]

        self.weights = np.zeros((self.diffusion_matrix.shape[1], 1))

        print(self.weights.shape)
        for k in range(len(self.table_index)):
            self.weights[self.table_index[k], 0] = self.table[weights_col].values.tolist()[k]

    def empirical_pvalue(self, geneset: list, alternative: str = "less", max_iter: int = 100, cores: int = 1) -> \
        [int, float, float, int, int]:
        """
        Calculate the empirical pvalue on the genes list

        :param geneset: the geneset to elaborate
        :param alternative: the pvalue selection of the observed genes
        :param max_iter: the number of iterations to be performed
        :param cores: the number of cores to be used
        :return observed, pvalue, null_distribution, len(mapped_genesetA), len(mapped_genesetB): the list with the data calculated
        """
        # mapping geneset

        geneset = geneset
        mapped_geneset = list(self.universe.intersection(set(geneset)))

        if len(mapped_geneset) == 0:
            return 0, 0, np.array([0]), 0, 0
        else:
            geneset_index = [map_table2matrix(self.nodes, i) for i in mapped_geneset]
            logging.info("Mapped %d genes out of %d." % (len(mapped_geneset), len(geneset)))

            observed = self.test_statistic(self.diffusion_matrix, self.weights, geneset_index, self.diz,
                                           observed_flag=True)
            logging.info("Observed %f." % observed)

            # iterations
            null_distribution = DiffusionTest.get_null_distribution_mp(self, geneset_index, max_iter, n_proc=cores)

            # computing empirical pvalue
            pvalue = 1
            if alternative == "greater":
                pvalue = (np.sum(null_distribution >= observed) + 1) / (float(max_iter) + 1)
            else:
                pvalue = (np.sum(null_distribution <= observed) + 1) / (float(max_iter) + 1)

            return observed, pvalue, null_distribution, len(mapped_geneset), len(geneset)

    def get_null_distribution_mp(self, geneset_index: list, iter: int = 100, n_proc: int = 1) -> np.ndarray:
        """
        Calculate the null distribution with multiple cores on the geneset

        :param geneset_index: the geneset id that point to the geneset to be used
        :param iter: the number of iterations to perform
        :param n_proc: the number of cpu to use for the elaboration
        :return: the array with null distribution
        """
        print("n_proc=" + str(n_proc))

        if n_proc == 1:
            null_distribution = DiffusionTest.get_null_distribution(self, geneset_index, iter)
        else:
            p = multiprocessing.Pool(n_proc)
            n_trial = int(iter / n_proc)
            print("n_trial=" + str(n_trial))
            results = [p.apply_async(DiffusionTest.get_null_distribution,
                                     args=(self, geneset_index, n_trial), ) for w in list(range(1, n_proc + 1))]
            null_distribution = np.array([])
            for r in results:
                null_distribution = np.hstack((null_distribution, np.array(r.get())))
            print(len(null_distribution))
            p.close()

        return np.asarray(null_distribution)

    def get_null_distribution(self, geneset_index: list, n_samples: int, randomize: str = "index") -> list:
        """
        Calculate the null distribution over the geneset

        :param geneset_index: the geneset id that points to the geneset to be used
        :param n_samples: the number of samples to be taken
        :return: the random distribution calculated for each element
        """
        np.random.seed()
        random_dist = []
        for i in range(n_samples):
            random_weights = self.weights.copy()
            if randomize != "index":
                np.random.shuffle(random_weights)
            else:
                np.random.shuffle(geneset_index)
            random_dist.append(self.test_statistic(self.diffusion_matrix, random_weights, geneset_index, self.diz,
                                                   observed_flag=True))
        return random_dist


###############################################################################
###  TEST STATISTICS FOR SINGLE GENESET  ######################################
###############################################################################


def weights_diffusion_statistic(matrix: np.matrix, weights: np.matrix, geneset_index: list, diz: dict = {},
                                observed_flag: bool = False) -> float:
    """
    Not in use.
    This statistic reweights the original weights and
    returns the average reweighted statistic.
    """

    if matrix.shape[1] != weights.shape[0]:
        logging.warning("pass the right shape for the weights")
    try:
        product = np.matmul(matrix, weights)
    except:
        logging.warning("error in matrix multiplication")

    w = [product[i] for i in geneset_index]
    stat = np.sum(w) / len(w)
    return stat


def hotnet_diffusion_statistic(matrix: np.matrix, weights: np.matrix, geneset_index: list, diz: dict = {},
                               observed_flag: bool = False) -> np.ndarray:
    """
    HOTNET2 like diffusion.
    Applies the diagonal matrix of weights and gets all rows and columns according to the genelist

    :param matrix: the matrix corresponding to the graph
    :param weights: the matrix of which it will be created the diagonal matrix
    :param geneset_index: the gene list index
    :param observed_flag: TBD
    """
    weights = np.diagflat(weights.T)
    if matrix.shape[1] != weights.shape[0]:
        logging.warning("pass the right shape for the weights")
    try:
        product = np.matmul(matrix, weights)
    except:
        logging.warning("error in matrix multiplication")

    prob = [product[i, j] for i in geneset_index for j in geneset_index if i != j]
    stat = np.sum(prob)
    return stat



from matplotlib.offsetbox import AnchoredText
from abc import ABC
import textwrap
import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from palettable.colorbrewer.diverging import *
from palettable.colorbrewer.sequential import *


class PygnaFigure(ABC):
    """
    Abstract class that implements a general figure in Pygna. It has a class attribute representing the default height/width ratios for the figures
    """
    ratios = {
        "volcano": [3, 4],
        "heatmap": [4, 3],
        "distribution": [4, 3],
        "barplot": [3, 4]
    }

    def __init__(self):
        super(PygnaFigure, self).__init__()
        self.filename = None

    def _save_fig(self) -> None:
        """
        This method saves the figure using the matplotlib framework
        """

        plt.savefig(self.filename)

    def _get_dimensions(self, fig_type: str, size: int) -> [int, int, int, int]:
        """
        This methods maps the ratios saved in the class with the figure type and returns the correct ration for each
        figure

        :param fig_type:  the figure type to be printed
        :param size:  the size of the plot
        :return: the width, height, fontsize, scalar of the figure
        """
        width = self.ratios[fig_type][0] * size
        height = self.ratios[fig_type][1] * size
        fontsize = size * 5
        if size <= 3:
            scalar = size ** 4
        else:
            scalar = size ** 3
        return int(width), int(height), int(fontsize), int(scalar)


class VolcanoPlot(PygnaFigure):
    """
    This class represent a Volcano Plot. It saves the value in the dataframe on a file.
    """

    def __init__(self, df: pd.DataFrame, output_file: str, loc: int = 2, p_col: str = "empirical_pvalue",
                 id_col: str = "setname_B", plotting_col: str = "observed", x_threshold: float = 0.1,
                 y_threshold: float = 0.1, y_label: str = "-log10(pvalue)", x_label: str = "z-score",
                 annotate: bool = False, size: int = 2):
        """
        :param df: the dataframe containing the data
        :param output_file: the output file
        :param loc: the location where to print the legend
        :param p_col: the column containing the empirical p-value values
        :param id_col: the column containing the setname
        :param plotting_col: the column to plot
        :param x_threshold: value of the threshold for the x axis
        :param y_threshold: value of the threshold for the y axis
        :param y_label: label for the y axis
        :param x_label: label for the x axis
        :param annotate: whether should be printed the annotaion table
        :param size: the size of the plots
        """
        super().__init__()
        self.df = df
        self.filename = output_file
        self.p_col = p_col
        self.id_col = id_col
        self.plotting_col = plotting_col
        self.x_thresh = x_threshold
        self.y_thresh = y_threshold
        self.y_label = y_label
        self.x_label = x_label
        self.annotate = annotate
        self.location = loc
        self.size = int(size)

        logging.info('Significant are considered when  %s > %f and %s > %f' % (self.p_col, self.y_thresh,
                                                                               self.plotting_col, self.x_thresh))
        self.df = self.df.sort_values(by=[p_col, plotting_col], ascending=True)
        not_sig = self._elaborate_not_sig_genes(self.df, self.plotting_col, self.p_col, self.y_thresh, self.x_thresh)
        sig = self._elaborate_sig_genes(self.df, self.plotting_col, self.p_col, self.y_thresh, self.x_thresh)
        text = [['Top 5 terms']]
        width, height, font_ratio, marker_scalar = super()._get_dimensions("volcano", self.size)
        fig, ax = plt.subplots(1, figsize=(width, height))

        for label in (ax.get_xticklabels() + ax.get_yticklabels()):
            label.set_fontsize(font_ratio)
        ax.scatter(not_sig[plotting_col], not_sig[p_col], marker="o", alpha=1, edgecolors=None,
                   color=RdBu_4.mpl_colors[3], s=marker_scalar)
        if len(sig) < 1:
            logging.error("There are no significant terms, not plotting the volcano plot")
        else:
            ax.scatter(sig[plotting_col], sig[p_col], marker="+", alpha=1, edgecolors=None, color=RdBu_4.mpl_colors[0],
                       s=marker_scalar)
            if len(sig) > 5:
                ax.scatter(sig.iloc[:5, :][plotting_col], sig.iloc[:5, :][p_col], marker="*", alpha=1, s=marker_scalar,
                           edgecolors=None, color=RdBu_4.mpl_colors[0])
                self._append_name(sig.iloc[:5, :].iterrows(), self.id_col, text)

        ax.axhline(y=self.y_thresh, xmin=0, xmax=1, alpha=0.5, color='k', linestyle='--', linewidth=0.5)
        ax.axvline(x=self.x_thresh, ymin=0, ymax=1, alpha=0.5, color='k', linestyle='--', linewidth=0.5)
        ax.set_xlabel(self.x_label, fontsize=font_ratio)
        ax.set_ylabel(self.y_label, fontsize=font_ratio)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        if self.annotate:
            texts = [sublist if type(sublist) == str else item for sublist in text for item in sublist]
            anchored_text = AnchoredText('\n'.join(texts), self.location, prop={'fontsize': font_ratio})
            ax.add_artist(anchored_text)

        super()._save_fig()

    def _elaborate_not_sig_genes(self, df, plotting_column, pvalue_col, y_thresh, x_thresh):
        """
        This method elaborates the genes which are not significant

        :param df: pd.datafrme, the dataframe to elaborate
        :param plotting_column: str, the column to apply the x threshold
        :param pvalue_col: str, the column to apply the y threshold
        :param y_thresh: float, the value of the x threshold
        :param x_thresh: float, the value of y threshold
        :return: pd.dataframe, the dataframe with the not significant genes
        """
        dataframe = df[(np.abs(df[plotting_column]) < x_thresh) | (df[pvalue_col] < y_thresh)].copy()
        return dataframe

    def _elaborate_sig_genes(self, df, plotting_column, pvalue_col, y_thresh, x_thresh):
        """
        This method elaborates the genes which are significant

        :param df: pd.datafrme, the dataframe to elaborate
        :param plotting_column: str, the column to apply the x threshold
        :param pvalue_col: str, the column to apply the y threshold
        :param y_thresh: float, the value of the x threshold
        :param x_thresh: float, the value of y threshold
        :return: pd.dataframe, the dataframe with the significant genes
        """
        dataframe = df[(np.abs(df[plotting_column]) >= x_thresh) & (df[pvalue_col] >= y_thresh)].copy()
        if len(dataframe) < 1:
            logging.warning("There are no significant terms")
        return dataframe

    def _append_name(self, dataset, id_col, text_list):
        """
        This method add the names to the legend of the plot

        :param dataset: df.dataframe, the dataset where are stored the names
        :param id_col: str, the column where the names are stored
        :param text_list: list, list with the names already stored
        :return: list, list with the names
        """
        for k, row in dataset:
            w = textwrap.wrap('-' + row[id_col].replace("_", " "), 30, break_long_words=False)
            text_list.append(w)
        return text_list
